import unittest

from pypika import (
    AliasedQuery,
    Case,
    MySQLQuery,
    PostgreSQLQuery,
    Query,
    Table,
    Tables,
)
from pypika import (
    Field as F,
)
from pypika import (
    functions as fn,
)
from pypika.functions import (
    Avg,
    Cast,
)
from pypika.terms import Values
from pypika.utils import QueryException

__author__ = "Timothy Heys"
__email__ = "theys@kayak.com"


class InsertIntoTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_insert_one_column(self):
        query = Query.into(self.table_abc).insert(1)

        self.assertEqual('INSERT INTO "abc" VALUES (1)', str(query))

    def test_insert_one_column_single_element_array(self):
        query = Query.into(self.table_abc).insert((1,))

        self.assertEqual('INSERT INTO "abc" VALUES (1)', str(query))

    def test_insert_one_column_multi_element_array(self):
        query = Query.into(self.table_abc).insert((1,), (2,))

        self.assertEqual('INSERT INTO "abc" VALUES (1),(2)', str(query))

    def test_insert_single_row_with_array_value(self):
        query = Query.into(self.table_abc).insert(1, ["a", "b", "c"])

        self.assertEqual("INSERT INTO \"abc\" VALUES (1,['a','b','c'])", str(query))

    def test_insert_multiple_rows_with_array_value(self):
        query = Query.into(self.table_abc).insert(
            (1, ["a", "b", "c"]),
            (2, ["c", "d", "e"]),
        )

        self.assertEqual(
            'INSERT INTO "abc" VALUES (1,[\'a\',\'b\',\'c\']),(2,[\'c\',\'d\',\'e\'])',
            str(query),
        )

    def test_insert_all_columns(self):
        query = Query.into(self.table_abc).insert(1, "a", True)

        self.assertEqual("INSERT INTO \"abc\" VALUES (1,'a',true)", str(query))

    def test_insert_all_columns_single_element(self):
        query = Query.into(self.table_abc).insert((1, "a", True))

        self.assertEqual("INSERT INTO \"abc\" VALUES (1,'a',true)", str(query))

    def test_insert_all_columns_multi_rows(self):
        query = Query.into(self.table_abc).insert((1, "a", True), (2, "b", False))

        self.assertEqual("INSERT INTO \"abc\" VALUES (1,'a',true),(2,'b',false)", str(query))

    def test_insert_all_columns_multi_rows_chained(self):
        query = Query.into(self.table_abc).insert(1, "a", True).insert(2, "b", False)

        self.assertEqual("INSERT INTO \"abc\" VALUES (1,'a',true),(2,'b',false)", str(query))

    def test_insert_all_columns_multi_rows_chained_mixed(self):
        query = Query.into(self.table_abc).insert((1, "a", True), (2, "b", False)).insert(3, "c", True)

        self.assertEqual(
            'INSERT INTO "abc" VALUES (1,\'a\',true),(2,\'b\',false),(3,\'c\',true)',
            str(query),
        )

    def test_insert_all_columns_multi_rows_chained_multiple_rows(self):
        query = (
            Query.into(self.table_abc).insert((1, "a", True), (2, "b", False)).insert((3, "c", True), (4, "d", False))
        )

        self.assertEqual(
            'INSERT INTO "abc" VALUES (1,\'a\',true),(2,\'b\',false),(3,\'c\',true),(4,\'d\',false)',
            str(query),
        )

    def test_insert_selected_columns(self):
        query = (
            Query.into(self.table_abc)
            .columns(self.table_abc.foo, self.table_abc.bar, self.table_abc.buz)
            .insert(1, "a", True)
        )

        self.assertEqual('INSERT INTO "abc" ("foo","bar","buz") VALUES (1,\'a\',true)', str(query))

    def test_insert_empty_columns(self):
        query = Query.into(self.table_abc).columns().insert(1, "a", True)

        self.assertEqual('INSERT INTO "abc" VALUES (1,\'a\',true)', str(query))

    def test_insert_selected_columns_type_list(self):
        query = (
            Query.into(self.table_abc)
            .columns([self.table_abc.foo, self.table_abc.bar, self.table_abc.buz])
            .insert(1, "a", True)
        )

        self.assertEqual('INSERT INTO "abc" ("foo","bar","buz") VALUES (1,\'a\',true)', str(query))

    def test_insert_empty_columns_type_list(self):
        query = Query.into(self.table_abc).columns([]).insert(1, "a", True)

        self.assertEqual('INSERT INTO "abc" VALUES (1,\'a\',true)', str(query))

    def test_insert_selected_columns_type_tuple(self):
        query = (
            Query.into(self.table_abc)
            .columns((self.table_abc.foo, self.table_abc.bar, self.table_abc.buz))
            .insert(1, "a", True)
        )

        self.assertEqual('INSERT INTO "abc" ("foo","bar","buz") VALUES (1,\'a\',true)', str(query))

    def test_insert_empty_columns_type_tuple(self):
        query = Query.into(self.table_abc).columns(()).insert(1, "a", True)

        self.assertEqual('INSERT INTO "abc" VALUES (1,\'a\',true)', str(query))

    def test_insert_none_skipped(self):
        query = Query.into(self.table_abc).insert()

        self.assertEqual("", str(query))

    def test_insert_ignore(self):
        query = Query.into(self.table_abc).insert(1).ignore()

        self.assertEqual('INSERT IGNORE INTO "abc" VALUES (1)', str(query))

    def test_insert_null(self):
        query = Query.into(self.table_abc).insert(None)

        self.assertEqual('INSERT INTO "abc" VALUES (NULL)', str(query))

    def test_insert_column_using_table_alias(self):
        q = self.table_abc.insert(1)
        self.assertEqual('INSERT INTO "abc" VALUES (1)', str(q))

    def test_insert_column_using_alias_with_chain(self):
        q = self.table_abc.insert(1, "a", True).insert(2, "b", False)
        self.assertEqual("INSERT INTO \"abc\" VALUES (1,'a',true),(2,'b',false)", str(q))

    def test_insert_with_statement(self):
        sub_query = Query().select(self.table_abc.id).from_(self.table_abc)
        aliased = AliasedQuery('sub_qs')

        q = Query().with_(sub_query, 'sub_qs').into(self.table_abc).select(aliased.id).from_(aliased)
        self.assertEqual(
            'WITH sub_qs AS (SELECT "id" FROM "abc") INSERT INTO "abc" SELECT "sub_qs"."id" FROM sub_qs',
            str(q),
        )


class PostgresInsertIntoOnConflictTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_insert_on_conflict_do_nothing_field(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1).on_conflict(self.table_abc.id).do_nothing()

        self.assertEqual('INSERT INTO "abc" VALUES (1) ON CONFLICT ("id") DO NOTHING', str(query))

    def test_insert_on_conflict_do_nothing_multiple_fields(self):
        query = (
            PostgreSQLQuery.into(self.table_abc)
            .insert(1)
            .on_conflict(self.table_abc.id, self.table_abc.sub_id)
            .do_nothing()
        )

        self.assertEqual('INSERT INTO "abc" VALUES (1) ON CONFLICT ("id", "sub_id") DO NOTHING', str(query))

    def test_insert_on_conflict_do_nothing_field_str(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1).on_conflict("id").do_nothing()

        self.assertEqual('INSERT INTO "abc" VALUES (1) ON CONFLICT ("id") DO NOTHING', str(query))

    def test_insert_on_conflict_do_nothing_multiple_fields_str(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1).on_conflict("id", "sub_id").do_nothing()

        self.assertEqual('INSERT INTO "abc" VALUES (1) ON CONFLICT ("id", "sub_id") DO NOTHING', str(query))

    def test_insert_on_conflict_do_nothing_mixed_fields(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1).on_conflict("id", self.table_abc.sub_id).do_nothing()

        self.assertEqual('INSERT INTO "abc" VALUES (1) ON CONFLICT ("id", "sub_id") DO NOTHING', str(query))

    def test_insert_on_conflict_do_update_field(self):
        query = (
            PostgreSQLQuery.into(self.table_abc)
            .insert(1, "m")
            .on_conflict(self.table_abc.id)
            .do_update(self.table_abc.name, "m")
        )

        self.assertEqual(
            "INSERT INTO \"abc\" VALUES (1,'m') ON CONFLICT (\"id\") DO UPDATE SET \"name\"='m'",
            str(query),
        )

    def test_insert_on_conflict_do_update_multiple_fields(self):
        query = (
            PostgreSQLQuery.into(self.table_abc)
            .insert(1, "m")
            .on_conflict(self.table_abc.id, self.table_abc.sub_id)
            .do_update(self.table_abc.name, "m")
        )

        self.assertEqual(
            "INSERT INTO \"abc\" VALUES (1,'m') ON CONFLICT (\"id\", \"sub_id\") DO UPDATE SET \"name\"='m'",
            str(query),
        )

    def test_insert_on_conflict_do_update_field_str(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1, "m").on_conflict("id").do_update("name", "m")

        self.assertEqual(
            "INSERT INTO \"abc\" VALUES (1,'m') ON CONFLICT (\"id\") DO UPDATE SET \"name\"='m'",
            str(query),
        )

    def test_insert_on_conflict_do_update_multiple_fields_str(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1, "m").on_conflict("id", "sub_id").do_update("name", "m")

        self.assertEqual(
            "INSERT INTO \"abc\" VALUES (1,'m') ON CONFLICT (\"id\", \"sub_id\") DO UPDATE SET \"name\"='m'",
            str(query),
        )

    def test_insert_on_conflict_do_update_multiple_mixed_fields(self):
        query = (
            PostgreSQLQuery.into(self.table_abc)
            .insert(1, "m")
            .on_conflict("id", self.table_abc.sub_id)
            .do_update("name", "m")
        )

        self.assertEqual(
            "INSERT INTO \"abc\" VALUES (1,'m') ON CONFLICT (\"id\", \"sub_id\") DO UPDATE SET \"name\"='m'",
            str(query),
        )

    def test_insert_on_conflict_no_handler(self):
        with self.assertRaises(QueryException):
            str(PostgreSQLQuery.into(self.table_abc).insert(1).on_conflict(self.table_abc.id))

    def test_insert_on_conflict_two_handlers_do_nothing(self):
        with self.assertRaises(QueryException):
            (
                PostgreSQLQuery.into(self.table_abc)
                .insert(1)
                .on_conflict("id")
                .do_nothing()
                .do_update(self.table_abc.name, "m")
            )

    def test_insert_on_conflict_two_handlers_do_update(self):
        with self.assertRaises(QueryException):
            (
                PostgreSQLQuery.into(self.table_abc)
                .insert(1)
                .on_conflict(self.table_abc.id)
                .do_update(self.table_abc.name, "m")
                .do_nothing()
            )

    def test_non_insert_on_conflict_do_nothing(self):
        with self.assertRaises(QueryException):
            PostgreSQLQuery.update(self.table_abc).set("foo", "bar").on_conflict("id").do_nothing()

    def test_non_insert_on_conflict_do_update(self):
        with self.assertRaises(QueryException):
            (PostgreSQLQuery.update(self.table_abc).set("foo", "bar").on_conflict("id").do_update(["name"], ["m"]))

    def test_insert_on_fieldless_conflict_do_nothing(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1).on_conflict(None).do_nothing()

        self.assertEqual('INSERT INTO "abc" VALUES (1) ON CONFLICT DO NOTHING', str(query))

    def test_insert_on_fieldless_conflict_do_update_field(self):
        with self.assertRaises(QueryException):
            str(
                PostgreSQLQuery.into(self.table_abc)
                .insert(1, "m")
                .on_conflict(None)
                .do_update(self.table_abc.name, "m"),
            )

    def test_on_conflict_from_subquery(self):
        table_bcd = Table('bcd')
        query = (
            PostgreSQLQuery.into(self.table_abc)
            .insert(self.table_abc.fname, self.table_abc.lname)
            .select(table_bcd.fname, table_bcd.lname)
            .from_(table_bcd)
            .on_conflict(self.table_abc.id, self.table_abc.sub_id)
            .do_update(self.table_abc.fname, 1)
            .do_update(self.table_abc.lname, table_bcd.lname)
            .do_update(
                self.table_abc.cname,
                Case().when(self.table_abc.cname.eq('cname'), 'new_name').else_(self.table_abc.cname),
            )
        )

        self.assertEqual(
            'INSERT INTO "abc" VALUES ("fname","lname") '
            'ON CONFLICT ("id", "sub_id") '
            'DO UPDATE SET "fname"=1,"lname"="bcd"."lname",'
            '"cname"=CASE WHEN "abc"."cname"=\'cname\' THEN \'new_name\' ELSE "abc"."cname" END',
            str(query),
        )

    def test_on_conflict_where_empty_conflict_fields_do_nothing(self):
        with self.assertRaises(QueryException):
            (
                PostgreSQLQuery.into(self.table_abc)
                .insert(1, "m")
                .on_conflict()
                .where(self.table_abc.abc.eq(0))
                .where(self.table_abc.cde.eq(0))
                .do_nothing()
            )

    def test_on_conflict_where_conflict_fields_do_nothing(self):
        qs = (
            PostgreSQLQuery.into(self.table_abc)
            .insert(1, "m")
            .on_conflict('id')
            .where(self.table_abc.abc.eq(0))
            .where(self.table_abc.cde.eq(0))
            .do_nothing()
        )
        self.assertEqual(
            '''INSERT INTO "abc" VALUES (1,'m') ON CONFLICT ("id") WHERE "abc"=0 AND "cde"=0 DO NOTHING''',
            str(qs),
        )

    def test_on_conflict_where_empty_conflict_fields_do_update(self):
        with self.assertRaises(QueryException):
            (
                PostgreSQLQuery.into(self.table_abc)
                .insert(1, "m")
                .on_conflict()
                .where(self.table_abc.abc.eq(0))
                .where(self.table_abc.cde.eq(0))
                .do_update('field', 'val')
            )

    def test_on_conflict_where_conflict_fields_do_update(self):
        qs = (
            PostgreSQLQuery.into(self.table_abc)
            .insert(1, "m")
            .on_conflict('id')
            .where(self.table_abc.abc.eq(0))
            .where(self.table_abc.cde.eq(0))
            .do_update('field', 'val')
        )
        self.assertEqual(
            '''INSERT INTO "abc" VALUES (1,'m') ON CONFLICT ("id") WHERE "abc"=0 AND "cde"=0 '''
            '''DO UPDATE SET "field"='val\'''',
            str(qs),
        )

    def test_where_and_on_conflict_where(self):
        table_bcd = Table('bcd')

        qs = (
            PostgreSQLQuery.into(self.table_abc)
            .select(table_bcd.abc)
            .from_(table_bcd)
            .where(table_bcd.abc.eq('1'))
            .on_conflict('id')
            .where(self.table_abc.abc.eq(0))
            .where(self.table_abc.cde.eq(0))
            .do_update('field', 'val')
        )

        self.assertEqual(
            'INSERT INTO "abc" SELECT "abc" FROM "bcd" WHERE "abc"=\'1\' '
            'ON CONFLICT ("id") WHERE "abc"=0 AND "cde"=0 DO UPDATE SET "field"=\'val\'',
            str(qs),
        )

    def test_on_conflict_do_nothing_where(self):
        with self.assertRaises(QueryException):
            (
                PostgreSQLQuery.into(self.table_abc)
                .insert(1, "m")
                .on_conflict()
                .do_nothing()
                .where(self.table_abc.abc.eq(1))
            )

    def test_empty_on_conflict_do_update_where(self):
        with self.assertRaises(QueryException):
            (
                PostgreSQLQuery.into(self.table_abc)
                .insert(1, "m")
                .on_conflict()
                .do_update('abc', 1)
                .where(self.table_abc.abc.eq(1))
            )

    def test_on_conflict_do_update_where(self):
        qs = (
            PostgreSQLQuery.into(self.table_abc)
            .insert(1, "m")
            .on_conflict("id")
            .do_update('abc', 1)
            .where(self.table_abc.abc.eq(1))
        )
        self.assertEqual(
            'INSERT INTO "abc" VALUES (1,\'m\') ON CONFLICT ("id") DO UPDATE SET "abc"=1 WHERE "abc"."abc"=1',
            str(qs),
        )

    def test_on_conflict_do_update_with_excluded_where(self):
        qs = (
            PostgreSQLQuery.into(self.table_abc)
            .insert(1, "m")
            .on_conflict("id")
            .do_update('abc')
            .where(self.table_abc.abc.eq(1))
        )
        self.assertEqual(
            'INSERT INTO "abc" VALUES (1,\'m\') ON CONFLICT ("id") DO UPDATE SET "abc"=EXCLUDED."abc" WHERE '
            '"abc"."abc"=1',
            str(qs),
        )

    def test_on_conflict_where_complex(self):
        table_bcd = Table('bcd')

        qs = (
            PostgreSQLQuery.into(self.table_abc)
            .select(table_bcd.abc)
            .from_(table_bcd)
            .where(table_bcd.abc.eq('1'))
            .on_conflict('id')
            .where(self.table_abc.abc.eq(0))
            .where(self.table_abc.cde.eq(0))
            .do_update('field', 'val')
            .where(self.table_abc.id.eq(2))
            .where(self.table_abc.sub_id.eq(3))
        )
        self.assertEqual(
            'INSERT INTO "abc" SELECT "abc" FROM "bcd" WHERE "abc"=\'1\' '
            'ON CONFLICT ("id") WHERE "abc"=0 AND "cde"=0 '
            'DO UPDATE SET "field"=\'val\' WHERE "abc"."id"=2 AND "abc"."sub_id"=3',
            str(qs),
        )


class PostgresInsertIntoReturningTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_insert_returning_one_field(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1).returning(self.table_abc.id)

        self.assertEqual('INSERT INTO "abc" VALUES (1) RETURNING "id"', str(query))

    def test_insert_returning_one_field_str(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1).returning("id")

        self.assertEqual('INSERT INTO "abc" VALUES (1) RETURNING "id"', str(query))

    def test_insert_returning_all_fields(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1).returning(self.table_abc.star)

        self.assertEqual('INSERT INTO "abc" VALUES (1) RETURNING *', str(query))

    def test_insert_returning_all_fields_and_arithmetics(self):
        query = (
            PostgreSQLQuery.into(self.table_abc)
            .insert(1)
            .returning(self.table_abc.star, self.table_abc.f1 + self.table_abc.f2)
        )

        self.assertEqual('INSERT INTO "abc" VALUES (1) RETURNING *,"f1"+"f2"', str(query))

    def test_insert_all_columns_multi_rows_chained_returning_star(self):
        query = (
            PostgreSQLQuery.into(self.table_abc)
            .insert(1, "a", True)
            .insert(2, "b", False)
            .returning(self.table_abc.star)
        )

        self.assertEqual(
            "INSERT INTO \"abc\" VALUES (1,'a',true),(2,'b',false) RETURNING *",
            str(query),
        )

    def test_insert_all_columns_multi_rows_chained_returning_star_and_id(self):
        query = (
            PostgreSQLQuery.into(self.table_abc)
            .insert(1, "a", True)
            .insert(2, "b", False)
            .returning(
                self.table_abc.name,
                self.table_abc.star,
                self.table_abc.id,
            )
        )

        self.assertEqual(
            "INSERT INTO \"abc\" VALUES (1,'a',true),(2,'b',false) RETURNING *",
            str(query),
        )

    def test_insert_all_columns_multi_rows_chained_returning_star_str(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1, "a", True).insert(2, "b", False).returning("*")

        self.assertEqual(
            "INSERT INTO \"abc\" VALUES (1,'a',true),(2,'b',false) RETURNING *",
            str(query),
        )

    def test_insert_all_columns_single_element_arrays(self):
        query = PostgreSQLQuery.into(self.table_abc).insert((1, "a", True)).returning(self.table_abc.star)

        self.assertEqual("INSERT INTO \"abc\" VALUES (1,'a',true) RETURNING *", str(query))

    def test_insert_returning_null(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1).returning(None)

        self.assertEqual('INSERT INTO "abc" VALUES (1) RETURNING NULL', str(query))

    def test_insert_returning_tuple(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1).returning((1, 2, 3))

        self.assertEqual('INSERT INTO "abc" VALUES (1) RETURNING (1,2,3)', str(query))

    def test_insert_returning_arithmetics(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1).returning(self.table_abc.f1 + self.table_abc.f2)

        self.assertEqual('INSERT INTO "abc" VALUES (1) RETURNING "f1"+"f2"', str(query))

    def test_insert_returning_functions(self):
        query = PostgreSQLQuery.into(self.table_abc).insert(1).returning(Cast(self.table_abc.f1, "int"))

        self.assertEqual('INSERT INTO "abc" VALUES (1) RETURNING CAST("f1" AS INT)', str(query))

    def test_insert_returning_aggregate(self):
        with self.assertRaises(QueryException):
            PostgreSQLQuery.into(self.table_abc).insert(1).returning(Avg(self.table_abc.views))

    def test_insert_returning_from_other_table(self):
        table_cba = Table("cba")
        with self.assertRaises(QueryException):
            PostgreSQLQuery.into(self.table_abc).insert(1).returning(table_cba.id)


class InsertIntoOnDuplicateTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_insert_one_column(self):
        query = (
            MySQLQuery.into(self.table_abc).insert(1).on_duplicate_key_update(self.table_abc.foo, self.table_abc.foo)
        )
        self.assertEqual(
            "INSERT INTO `abc` VALUES (1) ON DUPLICATE KEY UPDATE `foo`=`foo`",
            str(query),
        )

    def test_insert_one_column_using_values(self):
        query = (
            MySQLQuery.into(self.table_abc)
            .insert(1)
            .on_duplicate_key_update(self.table_abc.foo, Values(self.table_abc.foo))
        )
        self.assertEqual(
            "INSERT INTO `abc` VALUES (1) ON DUPLICATE KEY UPDATE `foo`=VALUES(`foo`)",
            str(query),
        )

    def test_insert_one_column_single_element_array(self):
        query = (
            MySQLQuery.into(self.table_abc).insert((1,)).on_duplicate_key_update(self.table_abc.foo, self.table_abc.foo)
        )

        self.assertEqual(
            "INSERT INTO `abc` VALUES (1) ON DUPLICATE KEY UPDATE `foo`=`foo`",
            str(query),
        )

    def test_insert_one_column_multi_element_array(self):
        query = (
            MySQLQuery.into(self.table_abc)
            .insert((1,), (2,))
            .on_duplicate_key_update(self.table_abc.foo, self.table_abc.foo)
        )

        self.assertEqual(
            "INSERT INTO `abc` VALUES (1),(2) ON DUPLICATE KEY UPDATE `foo`=`foo`",
            str(query),
        )

    def test_insert_multiple_columns_on_duplicate_update_one_with_same_value(self):
        query = (
            MySQLQuery.into(self.table_abc)
            .insert(1, "a")
            .on_duplicate_key_update(self.table_abc.bar, Values(self.table_abc.bar))
        )

        self.assertEqual(
            "INSERT INTO `abc` VALUES (1,'a') ON DUPLICATE KEY UPDATE `bar`=VALUES(`bar`)",
            str(query),
        )

    def test_insert_multiple_columns_on_duplicate_update_one_with_different_value(self):
        query = MySQLQuery.into(self.table_abc).insert(1, "a").on_duplicate_key_update(self.table_abc.bar, "b")

        self.assertEqual(
            "INSERT INTO `abc` VALUES (1,'a') ON DUPLICATE KEY UPDATE `bar`='b'",
            str(query),
        )

    def test_insert_multiple_columns_on_duplicate_update_one_with_expression(self):
        query = (
            MySQLQuery.into(self.table_abc).insert(1, 2).on_duplicate_key_update(self.table_abc.bar, 4 + F("bar"))
        )  # todo sql expression? not python

        self.assertEqual(
            "INSERT INTO `abc` VALUES (1,2) ON DUPLICATE KEY UPDATE `bar`=4+`bar`",
            str(query),
        )

    def test_insert_multiple_columns_on_duplicate_update_one_with_expression_using_original_field_value(
        self,
    ):
        query = (
            MySQLQuery.into(self.table_abc)
            .insert(1, "a")
            .on_duplicate_key_update(self.table_abc.bar, fn.Concat(self.table_abc.bar, "update"))
        )

        self.assertEqual(
            "INSERT INTO `abc` VALUES (1,'a') ON DUPLICATE KEY UPDATE `bar`=CONCAT(`bar`,'update')",
            str(query),
        )

    def test_insert_multiple_columns_on_duplicate_update_one_with_expression_using_values(
        self,
    ):
        query = (
            MySQLQuery.into(self.table_abc)
            .insert(1, "a")
            .on_duplicate_key_update(self.table_abc.bar, fn.Concat(Values(self.table_abc.bar), "update"))
        )

        self.assertEqual(
            "INSERT INTO `abc` VALUES (1,'a') ON DUPLICATE KEY UPDATE `bar`=CONCAT(VALUES(`bar`),'update')",
            str(query),
        )

    def test_insert_multiple_columns_on_duplicate_update_multiple(self):
        query = (
            MySQLQuery.into(self.table_abc)
            .insert(1, "a", "b")
            .on_duplicate_key_update(self.table_abc.bar, "b")
            .on_duplicate_key_update(self.table_abc.baz, "c")
        )

        self.assertEqual(
            "INSERT INTO `abc` VALUES (1,'a','b') ON DUPLICATE KEY UPDATE `bar`='b',`baz`='c'",
            str(query),
        )

    def test_insert_multi_rows_chained_mixed_on_duplicate_update_multiple(self):
        query = (
            MySQLQuery.into(self.table_abc)
            .insert((1, "a", True), (2, "b", False))
            .insert(3, "c", True)
            .on_duplicate_key_update(self.table_abc.foo, self.table_abc.foo)
            .on_duplicate_key_update(self.table_abc.bar, Values(self.table_abc.bar))
        )

        self.assertEqual(
            "INSERT INTO `abc` VALUES (1,'a',true),(2,'b',false),(3,'c',true) "
            "ON DUPLICATE KEY UPDATE `foo`=`foo`,`bar`=VALUES(`bar`)",
            str(query),
        )

    def test_insert_selected_columns_on_duplicate_update_one(self):
        query = (
            MySQLQuery.into(self.table_abc)
            .columns(self.table_abc.foo, self.table_abc.bar, self.table_abc.baz)
            .insert(1, "a", True)
            .on_duplicate_key_update(self.table_abc.baz, False)
        )

        self.assertEqual(
            "INSERT INTO `abc` (`foo`,`bar`,`baz`) VALUES (1,'a',true) ON DUPLICATE KEY UPDATE `baz`=false",
            str(query),
        )

    def test_insert_selected_columns_on_duplicate_update_multiple(self):
        query = (
            MySQLQuery.into(self.table_abc)
            .columns(self.table_abc.foo, self.table_abc.bar, self.table_abc.baz)
            .insert(1, "a", True)
            .on_duplicate_key_update(self.table_abc.baz, False)
            .on_duplicate_key_update(self.table_abc.bar, Values(self.table_abc.bar))
        )

        self.assertEqual(
            "INSERT INTO `abc` (`foo`,`bar`,`baz`) VALUES (1,'a',true) "
            "ON DUPLICATE KEY UPDATE `baz`=false,`bar`=VALUES(`bar`)",
            str(query),
        )

    def test_insert_none_skipped(self):
        query = MySQLQuery.into(self.table_abc).insert().on_duplicate_key_update(self.table_abc.baz, False)

        self.assertEqual("", str(query))

    def test_insert_ignore(self):
        query = MySQLQuery.into(self.table_abc).insert(1).ignore().on_duplicate_key_update(self.table_abc.baz, False)

        self.assertEqual(
            "INSERT IGNORE INTO `abc` VALUES (1) ON DUPLICATE KEY UPDATE `baz`=false",
            str(query),
        )


class InsertSelectFromTests(unittest.TestCase):
    table_abc, table_efg, table_hij = Tables("abc", "efg", "hij")

    def test_insert_star(self):
        query = Query.into(self.table_abc).from_(self.table_efg).select("*")

        self.assertEqual('INSERT INTO "abc" SELECT * FROM "efg"', str(query))

    def test_insert_ignore_star(self):
        query = Query.into(self.table_abc).from_(self.table_efg).select("*").ignore()

        self.assertEqual('INSERT IGNORE INTO "abc" SELECT * FROM "efg"', str(query))

    def test_insert_from_columns(self):
        query = (
            Query.into(self.table_abc)
            .from_(self.table_efg)
            .select(self.table_efg.fiz, self.table_efg.buz, self.table_efg.baz)
        )

        self.assertEqual('INSERT INTO "abc" SELECT "fiz","buz","baz" FROM "efg"', str(query))

    def test_insert_columns_from_star(self):
        query = (
            Query.into(self.table_abc)
            .columns(
                self.table_abc.foo,
                self.table_abc.bar,
                self.table_abc.buz,
            )
            .from_(self.table_efg)
            .select("*")
        )

        self.assertEqual('INSERT INTO "abc" ("foo","bar","buz") SELECT * FROM "efg"', str(query))

    def test_insert_columns_from_columns(self):
        query = (
            Query.into(self.table_abc)
            .columns(self.table_abc.foo, self.table_abc.bar, self.table_abc.buz)
            .from_(self.table_efg)
            .select(self.table_efg.fiz, self.table_efg.buz, self.table_efg.baz)
        )

        self.assertEqual(
            'INSERT INTO "abc" ("foo","bar","buz") SELECT "fiz","buz","baz" FROM "efg"',
            str(query),
        )

    def test_insert_columns_from_columns_with_join(self):
        query = (
            Query.into(self.table_abc)
            .columns(
                self.table_abc.c1,
                self.table_abc.c2,
                self.table_abc.c3,
                self.table_abc.c4,
            )
            .from_(self.table_efg)
            .select(self.table_efg.foo, self.table_efg.bar)
            .join(self.table_hij)
            .on(self.table_efg.id == self.table_hij.abc_id)
            .select(self.table_hij.fiz, self.table_hij.buz)
        )

        self.assertEqual(
            'INSERT INTO "abc" ("c1","c2","c3","c4") '
            'SELECT "efg"."foo","efg"."bar","hij"."fiz","hij"."buz" FROM "efg" '
            'JOIN "hij" ON "efg"."id"="hij"."abc_id"',
            str(query),
        )


class InsertSubqueryTests(unittest.TestCase):
    def test_insert_subquery_wrapped_in_brackets(self):
        purchase_order_item, part = Tables("purchase_order_item", "part")

        q = (
            Query.into(purchase_order_item)
            .columns(purchase_order_item.id_part, purchase_order_item.id_customer)
            .insert(
                Query.from_(part).select(part.part_id).where(part.part_number == "FOOBAR"),
                12345,
            )
        )

        self.assertEqual(
            'INSERT INTO "purchase_order_item" '
            '("id_part","id_customer") '
            "VALUES "
            '((SELECT "part_id" FROM "part" WHERE "part_number"=\'FOOBAR\'),12345)',
            str(q),
        )


class SelectIntoTests(unittest.TestCase):
    table_abc, table_efg, table_hij = Tables("abc", "efg", "hij")

    def test_select_star_into(self):
        query = Query.from_(self.table_abc).select("*").into(self.table_efg)

        self.assertEqual('SELECT * INTO "efg" FROM "abc"', str(query))

    def test_select_columns_into(self):
        query = (
            Query.from_(self.table_abc)
            .select(self.table_abc.foo, self.table_abc.bar, self.table_abc.buz)
            .into(self.table_efg)
        )

        self.assertEqual('SELECT "foo","bar","buz" INTO "efg" FROM "abc"', str(query))

    def test_select_columns_into_with_join(self):
        query = (
            Query.from_(self.table_abc)
            .select(self.table_abc.foo, self.table_abc.bar)
            .join(self.table_hij)
            .on(self.table_abc.id == self.table_hij.abc_id)
            .select(self.table_hij.fiz, self.table_hij.buz)
            .into(self.table_efg)
        )

        self.assertEqual(
            'SELECT "abc"."foo","abc"."bar","hij"."fiz","hij"."buz" '
            'INTO "efg" FROM "abc" '
            'JOIN "hij" ON "abc"."id"="hij"."abc_id"',
            str(query),
        )


class ReplaceTests(unittest.TestCase):
    table_abc, table_def = Tables("abc", "efg")

    def test_replace_simple(self):
        query = Query.into(self.table_abc).replace("v1", "v2", "v3")
        expected_output = "REPLACE INTO \"abc\" VALUES ('v1','v2','v3')"
        self.assertEqual(str(query), expected_output)

    def test_replace_subquery(self):
        query = Query.into(self.table_abc).replace(Query.from_(self.table_def).select("f1", "f2"))
        expected_output = 'REPLACE INTO "abc" VALUES ((SELECT "f1","f2" FROM "efg"))'
        self.assertEqual(str(query), expected_output)



import unittest

from pypika import Query, Table
from pypika.pseudocolumns import (
    ColumnValue,
    ObjectID,
    ObjectValue,
    RowID,
    RowNum,
    SysDate,
)
from pypika.terms import PseudoColumn


class PseudoColumnsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(PseudoColumnsTest, cls).setUpClass()
        cls.table1 = Table("table1")

    def test_column_value(self):
        self.assertEqual("COLUMN_VALUE", ColumnValue)

    def test_object_id(self):
        self.assertEqual("OBJECT_ID", ObjectID)

    def test_object_value(self):
        self.assertEqual("OBJECT_VALUE", ObjectValue)

    def test_row_id(self):
        self.assertEqual("ROWID", RowID)

    def test_row_num(self):
        self.assertEqual("ROWNUM", RowNum)

    def test_sys_date(self):
        self.assertEqual("SYSDATE", SysDate)

    def test_can_be_used_in_a_select_statement(self):
        query = Query.from_(self.table1).where(self.table1.is_active == 1).select(PseudoColumn("abcde"))

        self.assertEqual(str(query), 'SELECT abcde FROM "table1" WHERE "is_active"=1')

    def test_can_be_used_in_a_where_clause(self):
        query = Query.from_(self.table1).where(PseudoColumn("abcde") > 1).select(self.table1.is_active)

        self.assertEqual(str(query), 'SELECT "is_active" FROM "table1" WHERE abcde>1')

    def test_can_be_used_in_orderby(self):
        query = Query.from_(self.table1).select(self.table1.abcde.as_("some_name")).orderby(PseudoColumn("some_name"))
        self.assertEqual(str(query), 'SELECT "abcde" "some_name" FROM "table1" ORDER BY some_name')



import unittest
from datetime import (
    date,
    datetime,
)

from pypika import (
    Criterion,
    EmptyCriterion,
    Field,
    Table,
)
from pypika import (
    functions as fn,
)
from pypika.queries import QueryBuilder
from pypika.terms import ExistsCriterion, Mod

__author__ = "Timothy Heys"
__email__ = "theys@kayak.com"


class CriterionTests(unittest.TestCase):
    t = Table("test", alias="crit")

    def test__criterion_with_alias(self):
        c1 = (Field("foo") == Field("bar")).as_("criterion")

        self.assertEqual('"foo"="bar"', str(c1))
        self.assertEqual('"foo"="bar" "criterion"', c1.get_sql(with_alias=True, quote_char='"', alias_quote_char='"'))

    def test__criterion_eq_number(self):
        c1 = Field("foo") == 1
        c2 = Field("foo", table=self.t).eq(0)
        c3 = Field("foo", table=self.t) == -1

        self.assertEqual('"foo"=1', str(c1))
        self.assertEqual('"crit"."foo"=0', str(c2))
        self.assertEqual('"crit"."foo"=-1', str(c3))

    def test__criterion_eq_decimal(self):
        c1 = Field("foo") == 1.0
        c2 = Field("foo", table=self.t).eq(0.5)

        self.assertEqual('"foo"=1.0', str(c1))
        self.assertEqual('"crit"."foo"=0.5', str(c2))

    def test__criterion_eq_bool(self):
        c1 = Field("foo") == True
        c2 = Field("foo", table=self.t).eq(True)
        c3 = Field("foo") == False
        c4 = Field("foo", table=self.t).eq(False)

        self.assertEqual('"foo"=true', str(c1))
        self.assertEqual('"crit"."foo"=true', str(c2))
        self.assertEqual('"foo"=false', str(c3))
        self.assertEqual('"crit"."foo"=false', str(c4))

    def test__criterion_eq_str(self):
        c1 = Field("foo") == "abc"
        c2 = Field("foo", table=self.t).eq("abc")

        self.assertEqual("\"foo\"='abc'", str(c1))
        self.assertEqual('"crit"."foo"=\'abc\'', str(c2))

    def test__criterion_eq_date(self):
        c1 = Field("foo") == date(2000, 1, 1)
        c2 = Field("foo", table=self.t).eq(date(2000, 1, 1))

        self.assertEqual("\"foo\"='2000-01-01'", str(c1))
        self.assertEqual('"crit"."foo"=\'2000-01-01\'', str(c2))

    def test__criterion_eq_datetime(self):
        c1 = Field("foo") == datetime(2000, 1, 1, 12, 30, 55)
        c2 = Field("foo", table=self.t).eq(datetime(2000, 1, 1, 12, 30, 55))

        self.assertEqual("\"foo\"='2000-01-01T12:30:55'", str(c1))
        self.assertEqual('"crit"."foo"=\'2000-01-01T12:30:55\'', str(c2))

    def test__criterion_eq_right(self):
        c1 = Field("foo") == 1
        c2 = Field("foo", table=self.t) == -1

        self.assertEqual('"foo"=1', str(c1))
        self.assertEqual('"crit"."foo"=-1', str(c2))

    def test__criterion_is_null(self):
        c1 = Field("foo").isnull()
        c2 = Field("foo", table=self.t).isnull()

        self.assertEqual('"foo" IS NULL', str(c1))
        self.assertEqual('"crit"."foo" IS NULL', str(c2))

    def test__criterion_is_null_with_alias(self):
        c1 = Field("foo").isnull().as_('alias')
        c2 = Field("foo", table=self.t).isnull().as_('alias')

        self.assertEqual('"foo" IS NULL "alias"', str(c1))
        self.assertEqual('"crit"."foo" IS NULL "alias"', str(c2))

    def test__criterion_ne_number(self):
        c1 = Field("foo") != 1
        c2 = Field("foo", table=self.t).ne(0)
        c3 = Field("foo") != -1

        self.assertEqual('"foo"<>1', str(c1))
        self.assertEqual('"crit"."foo"<>0', str(c2))
        self.assertEqual('"foo"<>-1', str(c3))

    def test__criterion_ne_str(self):
        c1 = Field("foo") != "abc"
        c2 = Field("foo", table=self.t).ne("abc")

        self.assertEqual("\"foo\"<>'abc'", str(c1))
        self.assertEqual('"crit"."foo"<>\'abc\'', str(c2))

    def test__criterion_ne_date(self):
        c1 = Field("foo") != date(2000, 1, 1)
        c2 = Field("foo", table=self.t).ne(date(2000, 1, 1))

        self.assertEqual("\"foo\"<>'2000-01-01'", str(c1))
        self.assertEqual('"crit"."foo"<>\'2000-01-01\'', str(c2))

    def test__criterion_ne_datetime(self):
        c1 = Field("foo") != datetime(2000, 1, 1, 12, 30, 55)
        c2 = Field("foo", table=self.t).ne(datetime(2000, 1, 1, 12, 30, 55))

        self.assertEqual("\"foo\"<>'2000-01-01T12:30:55'", str(c1))
        self.assertEqual('"crit"."foo"<>\'2000-01-01T12:30:55\'', str(c2))

    def test__criterion_ne_right(self):
        c1 = Field("foo") != 1
        c2 = Field("foo", table=self.t) != -1

        self.assertEqual('"foo"<>1', str(c1))
        self.assertEqual('"crit"."foo"<>-1', str(c2))

    def test__criterion_lt_number(self):
        c1 = Field("foo") < 1
        c2 = Field("foo", table=self.t).lt(0)
        c3 = Field("foo") < -1

        self.assertEqual('"foo"<1', str(c1))
        self.assertEqual('"crit"."foo"<0', str(c2))
        self.assertEqual('"foo"<-1', str(c3))

    def test__criterion_lt_date(self):
        c1 = Field("foo") < date(2000, 1, 1)
        c2 = Field("foo", table=self.t).lt(date(2000, 1, 1))

        self.assertEqual("\"foo\"<'2000-01-01'", str(c1))
        self.assertEqual('"crit"."foo"<\'2000-01-01\'', str(c2))

    def test__criterion_lt_datetime(self):
        c1 = Field("foo") < datetime(2000, 1, 1, 12, 30, 55)
        c2 = Field("foo", table=self.t).lt(datetime(2000, 1, 1, 12, 30, 55))

        self.assertEqual("\"foo\"<'2000-01-01T12:30:55'", str(c1))
        self.assertEqual('"crit"."foo"<\'2000-01-01T12:30:55\'', str(c2))

    def test__criterion_lt_right(self):
        c1 = Field("foo") < 1
        c2 = Field("foo", table=self.t) < -1

        self.assertEqual('"foo"<1', str(c1))
        self.assertEqual('"crit"."foo"<-1', str(c2))

    def test__criterion_gt_number(self):
        c1 = Field("foo") > 1
        c2 = Field("foo", table=self.t).gt(0)
        c3 = Field("foo") > -1

        self.assertEqual('"foo">1', str(c1))
        self.assertEqual('"crit"."foo">0', str(c2))
        self.assertEqual('"foo">-1', str(c3))

    def test__criterion_gt_date(self):
        c1 = Field("foo") > date(2000, 1, 1)
        c2 = Field("foo", table=self.t).gt(date(2000, 1, 1))

        self.assertEqual("\"foo\">'2000-01-01'", str(c1))
        self.assertEqual('"crit"."foo">\'2000-01-01\'', str(c2))

    def test__criterion_gt_datetime(self):
        c1 = Field("foo") > datetime(2000, 1, 1, 12, 30, 55)
        c2 = Field("foo", table=self.t).gt(datetime(2000, 1, 1, 12, 30, 55))

        self.assertEqual("\"foo\">'2000-01-01T12:30:55'", str(c1))
        self.assertEqual('"crit"."foo">\'2000-01-01T12:30:55\'', str(c2))

    def test__criterion_gt_right(self):
        c1 = Field("foo") > 1
        c2 = Field("foo", table=self.t) > -1

        self.assertEqual('"foo">1', str(c1))
        self.assertEqual('"crit"."foo">-1', str(c2))

    def test__criterion_lte_number(self):
        c1 = Field("foo") <= 1
        c2 = Field("foo", table=self.t).lte(0)
        c3 = Field("foo") <= -1

        self.assertEqual('"foo"<=1', str(c1))
        self.assertEqual('"crit"."foo"<=0', str(c2))
        self.assertEqual('"foo"<=-1', str(c3))

    def test__criterion_lte_date(self):
        c1 = Field("foo") <= date(2000, 1, 1)
        c2 = Field("foo", table=self.t).lte(date(2000, 1, 1))

        self.assertEqual("\"foo\"<='2000-01-01'", str(c1))
        self.assertEqual('"crit"."foo"<=\'2000-01-01\'', str(c2))

    def test__criterion_lte_datetime(self):
        c1 = Field("foo") <= datetime(2000, 1, 1, 12, 30, 55)
        c2 = Field("foo", table=self.t).lte(datetime(2000, 1, 1, 12, 30, 55))

        self.assertEqual("\"foo\"<='2000-01-01T12:30:55'", str(c1))
        self.assertEqual('"crit"."foo"<=\'2000-01-01T12:30:55\'', str(c2))

    def test__criterion_lte_right(self):
        c1 = Field("foo") <= 1
        c2 = Field("foo", table=self.t) <= -1

        self.assertEqual('"foo"<=1', str(c1))
        self.assertEqual('"crit"."foo"<=-1', str(c2))

    def test__criterion_gte_number(self):
        c1 = Field("foo") >= 1
        c2 = Field("foo", table=self.t).gte(0)
        c3 = Field("foo") >= -1

        self.assertEqual('"foo">=1', str(c1))
        self.assertEqual('"crit"."foo">=0', str(c2))
        self.assertEqual('"foo">=-1', str(c3))

    def test__criterion_gte_date(self):
        c1 = Field("foo") >= date(2000, 1, 1)
        c2 = Field("foo", table=self.t).gte(date(2000, 1, 1))

        self.assertEqual("\"foo\">='2000-01-01'", str(c1))
        self.assertEqual('"crit"."foo">=\'2000-01-01\'', str(c2))

    def test__criterion_gte_datetime(self):
        c1 = Field("foo") >= datetime(2000, 1, 1, 12, 30, 55)
        c2 = Field("foo", table=self.t).gte(datetime(2000, 1, 1, 12, 30, 55))

        self.assertEqual("\"foo\">='2000-01-01T12:30:55'", str(c1))
        self.assertEqual('"crit"."foo">=\'2000-01-01T12:30:55\'', str(c2))

    def test__criterion_gte_right(self):
        c1 = Field("foo") >= 1
        c2 = Field("foo", table=self.t) >= -1

        self.assertEqual('"foo">=1', str(c1))
        self.assertEqual('"crit"."foo">=-1', str(c2))

    def test__criterion_bitwise_and(self):
        c1 = Field("foo").bitwiseand(2)
        c2 = Field("foo", table=self.t).bitwiseand(10) == 2

        self.assertEqual('("foo" & 2)', str(c1))
        self.assertEqual('("crit"."foo" & 10)=2', str(c2))

    def test__criterion_bitwise_and_with_alias(self):
        c1 = Field("foo").bitwiseand(2).as_('alias')

        self.assertEqual('("foo" & 2) "alias"', str(c1))

    def test__bitwise_and_in_where_clause(self):
        q = QueryBuilder().from_('items').select('abc').where(Field("foo").bitwiseand(1) == 1)

        self.assertEqual('SELECT "abc" FROM "items" WHERE ("foo" & 1)=1', str(q))


class NotTests(unittest.TestCase):
    table_abc, table_efg = Table("abc", alias="cx0"), Table("efg", alias="cx1")

    def test_negate(self):
        c1 = Field("foo") >= 1
        c2 = c1.negate()

        self.assertEqual('"foo">=1', str(c1))
        self.assertEqual('NOT "foo">=1', str(c2))

    def test_variable_access(self):
        c1 = Field("foo").negate()

        self.assertEqual(c1.is_aggregate, False)

    def test_chained_function(self):
        field1 = Field("foo").negate()
        field2 = field1.eq("bar")

        self.assertEqual('NOT "foo"', str(field1))
        self.assertEqual("NOT \"foo\"='bar'", str(field2))
        self.assertIsNot(field1, field2)

    def test_not_null(self):
        c1 = Field("foo").notnull()
        c2 = Field("foo", table=self.table_abc).notnull()

        self.assertEqual('NOT "foo" IS NULL', str(c1))
        self.assertEqual('NOT "cx0"."foo" IS NULL', str(c2))

    def test_is_not_null(self):
        c1 = Field("foo").isnotnull()
        c2 = Field("foo", table=self.table_abc).isnotnull()

        self.assertEqual('"foo" IS NOT NULL', str(c1))
        self.assertEqual('"cx0"."foo" IS NOT NULL', str(c2))

    def test_not_null_with_alias(self):
        c1 = Field("foo").notnull().as_("something")
        c2 = Field("foo", table=self.table_abc).notnull().as_("something")

        self.assertEqual('NOT "foo" IS NULL "something"', str(c1))
        self.assertEqual('NOT "cx0"."foo" IS NULL "something"', str(c2))

    def test_notnullcriterion_replace_table(self):
        f = self.table_abc.foo.notnull().replace_table(self.table_abc, self.table_efg)

        self.assertEqual('NOT "cx1"."foo" IS NULL', str(f))

    def test_not_with_or_criterion(self):
        self.assertEqual('NOT ("foo" OR "bar")', str(~(Field("foo") | Field("bar"))))

    def test_not_with_and_criterion(self):
        self.assertEqual('NOT ("foo" AND "bar")', str(~(Field("foo") & Field("bar"))))

    def test_not_with_complex_criterion(self):
        self.assertEqual(
            'NOT ("foo" AND "bar" AND "fizz" AND "buzz")',
            str(~(Field("foo") & Field("bar") & Field("fizz") & Field("buzz"))),
        )


class BetweenTests(unittest.TestCase):
    t = Table("abc", alias="btw")

    def test__between_number(self):
        c1 = Field("foo").between(0, 1)
        c2 = Field("foo", table=self.t).between(0, 1)
        c3 = Field("foo")[0:1]

        self.assertEqual('"foo" BETWEEN 0 AND 1', str(c1))
        self.assertEqual('"btw"."foo" BETWEEN 0 AND 1', str(c2))
        self.assertEqual('"foo" BETWEEN 0 AND 1', str(c3))

    def test__between_with_alias(self):
        c1 = Field("foo").between(0, 1).as_('alias')
        c2 = Field("foo", table=self.t).between(0, 1).as_('alias')
        c3 = Field("foo")[0:1].as_('alias')

        self.assertEqual('"foo" BETWEEN 0 AND 1 "alias"', str(c1))
        self.assertEqual('"btw"."foo" BETWEEN 0 AND 1 "alias"', str(c2))
        self.assertEqual('"foo" BETWEEN 0 AND 1 "alias"', str(c3))

    def test__between_date(self):
        c1 = Field("foo").between(date(2000, 1, 1), date(2000, 12, 31))
        c2 = Field("foo", table=self.t).between(date(2000, 1, 1), date(2000, 12, 31))
        c3 = Field("foo")[date(2000, 1, 1) : date(2000, 12, 31)]

        self.assertEqual("\"foo\" BETWEEN '2000-01-01' AND '2000-12-31'", str(c1))
        self.assertEqual("\"btw\".\"foo\" BETWEEN '2000-01-01' AND '2000-12-31'", str(c2))
        self.assertEqual("\"foo\" BETWEEN '2000-01-01' AND '2000-12-31'", str(c3))

    def test__between_datetime(self):
        c1 = Field("foo").between(datetime(2000, 1, 1, 0, 0, 0), datetime(2000, 12, 31, 23, 59, 59))
        c2 = Field("foo", table=self.t).between(datetime(2000, 1, 1, 0, 0, 0), datetime(2000, 12, 31, 23, 59, 59))
        c3 = Field("foo")[datetime(2000, 1, 1, 0, 0, 0) : datetime(2000, 12, 31, 23, 59, 59)]

        self.assertEqual("\"foo\" BETWEEN '2000-01-01T00:00:00' AND '2000-12-31T23:59:59'", str(c1))
        self.assertEqual(
            "\"btw\".\"foo\" BETWEEN '2000-01-01T00:00:00' AND '2000-12-31T23:59:59'",
            str(c2),
        )
        self.assertEqual("\"foo\" BETWEEN '2000-01-01T00:00:00' AND '2000-12-31T23:59:59'", str(c3))

    def test__function_between(self):
        c1 = fn.Coalesce(Field("foo"), 0)[0:1]
        c2 = fn.Coalesce(Field("foo", table=self.t), 0)[0:1]

        self.assertEqual('COALESCE("foo",0) BETWEEN 0 AND 1', str(c1))
        self.assertEqual('COALESCE("btw"."foo",0) BETWEEN 0 AND 1', str(c2))

    def test_get_item_only_works_with_slice(self):
        with self.assertRaises(TypeError):
            Field("foo")[0]

        with self.assertRaises(TypeError):
            Field("foo")[date(2000, 1, 1)]

        with self.assertRaises(TypeError):
            Field("foo")[datetime(2000, 1, 1, 0, 0, 0)]


class IsInTests(unittest.TestCase):
    t = Table("abc", alias="isin")

    def test__in_number(self):
        c1 = Field("foo").isin([0, 1])
        c2 = Field("foo", table=self.t).isin([0, 1])

        self.assertEqual('"foo" IN (0,1)', str(c1))
        self.assertEqual('"isin"."foo" IN (0,1)', str(c2))

    def test__in_number_with_alias(self):
        c1 = Field("foo").isin([0, 1]).as_('alias')
        c2 = Field("foo", table=self.t).isin([0, 1]).as_('alias')

        self.assertEqual('"foo" IN (0,1) "alias"', str(c1))
        self.assertEqual('"isin"."foo" IN (0,1) "alias"', str(c2))

    def test__in_character(self):
        c1 = Field("foo").isin(["a", "b"])
        c2 = Field("foo", table=self.t).isin(["a", "b"])

        self.assertEqual("\"foo\" IN ('a','b')", str(c1))
        self.assertEqual("\"isin\".\"foo\" IN ('a','b')", str(c2))

    def test__in_date(self):
        c1 = Field("foo").isin([date(2000, 1, 1), date(2000, 12, 31)])
        c2 = Field("foo", table=self.t).isin([date(2000, 1, 1), date(2000, 12, 31)])

        self.assertEqual("\"foo\" IN ('2000-01-01','2000-12-31')", str(c1))
        self.assertEqual("\"isin\".\"foo\" IN ('2000-01-01','2000-12-31')", str(c2))

    def test__in_datetime(self):
        c1 = Field("foo").isin([datetime(2000, 1, 1, 0, 0, 0), datetime(2000, 12, 31, 23, 59, 59)])
        c2 = Field("foo", table=self.t).isin([datetime(2000, 1, 1, 0, 0, 0), datetime(2000, 12, 31, 23, 59, 59)])

        self.assertEqual("\"foo\" IN ('2000-01-01T00:00:00','2000-12-31T23:59:59')", str(c1))
        self.assertEqual("\"isin\".\"foo\" IN ('2000-01-01T00:00:00','2000-12-31T23:59:59')", str(c2))

    def test__function_isin(self):
        c1 = fn.Coalesce(Field("foo"), 0).isin([0, 1])
        c2 = fn.Coalesce(Field("foo", table=self.t), 0).isin([0, 1])

        self.assertEqual('COALESCE("foo",0) IN (0,1)', str(c1))
        self.assertEqual('COALESCE("isin"."foo",0) IN (0,1)', str(c2))

        for t in (list, tuple, set, frozenset):
            c_type = fn.Coalesce(Field("foo"), 0).isin(t([0, 1]))
            self.assertEqual('COALESCE("foo",0) IN (0,1)', str(c_type))

        self.assertRaises(AttributeError, lambda: str(fn.Coalesce(Field("foo"), 0).isin('SHOULD_FAIL')))

    def test__in_unicode(self):
        c1 = Field("foo").isin(["a", "b"])
        c2 = Field("foo", table=self.t).isin(["a", "b"])

        self.assertEqual("\"foo\" IN ('a','b')", str(c1))
        self.assertEqual("\"isin\".\"foo\" IN ('a','b')", str(c2))


class ArithmeticExpressionTests(unittest.TestCase):
    def test__lshift(self):
        c1 = Field("foo").lshift("1")
        c2 = Field("foo").lshift("2")

        self.assertEqual("\"foo\"<<'1'", str(c1))
        self.assertEqual("\"foo\"<<'2'", str(c2))

    def test__rshift(self):
        c1 = Field("foo").rshift("1")
        c2 = Field("foo").rshift("2")

        self.assertEqual("\"foo\">>'1'", str(c1))
        self.assertEqual("\"foo\">>'2'", str(c2))


class NotInTests(unittest.TestCase):
    t = Table("abc", alias="notin")

    def test__notin_number(self):
        c1 = Field("foo").notin([0, 1])
        c2 = Field("foo", table=self.t).notin([0, 1])

        self.assertEqual('"foo" NOT IN (0,1)', str(c1))
        self.assertEqual('"notin"."foo" NOT IN (0,1)', str(c2))

    def test__notin_number_with_alias(self):
        c1 = Field("foo").notin([0, 1]).as_('alias')
        c2 = Field("foo", table=self.t).notin([0, 1]).as_('alias')

        self.assertEqual('"foo" NOT IN (0,1) "alias"', str(c1))
        self.assertEqual('"notin"."foo" NOT IN (0,1) "alias"', str(c2))

    def test__notin_character(self):
        c1 = Field("foo").notin(["a", "b"])
        c2 = Field("foo", table=self.t).notin(["a", "b"])

        self.assertEqual("\"foo\" NOT IN ('a','b')", str(c1))
        self.assertEqual("\"notin\".\"foo\" NOT IN ('a','b')", str(c2))

    def test__notin_date(self):
        c1 = Field("foo").notin([date(2000, 1, 1), date(2000, 12, 31)])
        c2 = Field("foo", table=self.t).notin([date(2000, 1, 1), date(2000, 12, 31)])

        self.assertEqual("\"foo\" NOT IN ('2000-01-01','2000-12-31')", str(c1))
        self.assertEqual("\"notin\".\"foo\" NOT IN ('2000-01-01','2000-12-31')", str(c2))

    def test__notin_datetime(self):
        c1 = Field("foo").notin([datetime(2000, 1, 1, 0, 0, 0), datetime(2000, 12, 31, 23, 59, 59)])
        c2 = Field("foo", table=self.t).notin([datetime(2000, 1, 1, 0, 0, 0), datetime(2000, 12, 31, 23, 59, 59)])

        self.assertEqual("\"foo\" NOT IN ('2000-01-01T00:00:00','2000-12-31T23:59:59')", str(c1))
        self.assertEqual(
            "\"notin\".\"foo\" NOT IN ('2000-01-01T00:00:00','2000-12-31T23:59:59')",
            str(c2),
        )

    def test__function_notin(self):
        c1 = fn.Coalesce(Field("foo"), 0).notin([0, 1])
        c2 = fn.Coalesce(Field("foo", table=self.t), 0).notin([0, 1])

        self.assertEqual('COALESCE("foo",0) NOT IN (0,1)', str(c1))
        self.assertEqual('COALESCE("notin"."foo",0) NOT IN (0,1)', str(c2))


class LikeTests(unittest.TestCase):
    t = Table("abc", alias="like")

    def test_like_starts_with(self):
        c1 = Field("foo").like("ab%")
        c2 = Field("foo", table=self.t).like("ab%")

        self.assertEqual("\"foo\" LIKE 'ab%'", str(c1))
        self.assertEqual('"like"."foo" LIKE \'ab%\'', str(c2))

    def test_like_contains(self):
        c1 = Field("foo").like("%ab%")
        c2 = Field("foo", table=self.t).like("%ab%")

        self.assertEqual("\"foo\" LIKE '%ab%'", str(c1))
        self.assertEqual('"like"."foo" LIKE \'%ab%\'', str(c2))

    def test_like_ends_with(self):
        c1 = Field("foo").like("%ab")
        c2 = Field("foo", table=self.t).like("%ab")

        self.assertEqual("\"foo\" LIKE '%ab'", str(c1))
        self.assertEqual('"like"."foo" LIKE \'%ab\'', str(c2))

    def test_like_n_chars_long(self):
        c1 = Field("foo").like("___")
        c2 = Field("foo", table=self.t).like("___")

        self.assertEqual("\"foo\" LIKE '___'", str(c1))
        self.assertEqual('"like"."foo" LIKE \'___\'', str(c2))

    def test_like_single_chars_and_various_chars(self):
        c1 = Field("foo").like("a_b%c")
        c2 = Field("foo", table=self.t).like("a_b%c")

        self.assertEqual("\"foo\" LIKE 'a_b%c'", str(c1))
        self.assertEqual('"like"."foo" LIKE \'a_b%c\'', str(c2))

    def test_not_like_single_chars_and_various_chars(self):
        c1 = Field("foo").not_like("a_b%c")
        c2 = Field("foo", table=self.t).not_like("a_b%c")

        self.assertEqual("\"foo\" NOT LIKE 'a_b%c'", str(c1))
        self.assertEqual('"like"."foo" NOT LIKE \'a_b%c\'', str(c2))

    def test_ilike_starts_with(self):
        c1 = Field("foo").ilike("ab%")
        c2 = Field("foo", table=self.t).ilike("ab%")

        self.assertEqual("\"foo\" ILIKE 'ab%'", str(c1))
        self.assertEqual('"like"."foo" ILIKE \'ab%\'', str(c2))

    def test_ilike_contains(self):
        c1 = Field("foo").ilike("%ab%")
        c2 = Field("foo", table=self.t).ilike("%ab%")

        self.assertEqual("\"foo\" ILIKE '%ab%'", str(c1))
        self.assertEqual('"like"."foo" ILIKE \'%ab%\'', str(c2))

    def test_ilike_ends_with(self):
        c1 = Field("foo").ilike("%ab")
        c2 = Field("foo", table=self.t).ilike("%ab")

        self.assertEqual("\"foo\" ILIKE '%ab'", str(c1))
        self.assertEqual('"like"."foo" ILIKE \'%ab\'', str(c2))

    def test_ilike_n_chars_long(self):
        c1 = Field("foo").ilike("___")
        c2 = Field("foo", table=self.t).ilike("___")

        self.assertEqual("\"foo\" ILIKE '___'", str(c1))
        self.assertEqual('"like"."foo" ILIKE \'___\'', str(c2))

    def test_ilike_single_chars_and_various_chars(self):
        c1 = Field("foo").ilike("a_b%c")
        c2 = Field("foo", table=self.t).ilike("a_b%c")

        self.assertEqual("\"foo\" ILIKE 'a_b%c'", str(c1))
        self.assertEqual('"like"."foo" ILIKE \'a_b%c\'', str(c2))

    def test_not_ilike_single_chars_and_various_chars(self):
        c1 = Field("foo").not_ilike("a_b%c")
        c2 = Field("foo", table=self.t).not_ilike("a_b%c")

        self.assertEqual("\"foo\" NOT ILIKE 'a_b%c'", str(c1))
        self.assertEqual('"like"."foo" NOT ILIKE \'a_b%c\'', str(c2))

    def test_rlike_escape_chars(self):
        c1 = Field("foo").rlike("\\\\d+$")
        c2 = Field("foo", table=self.t).rlike("\\\\d+$")

        self.assertEqual("\"foo\" RLIKE '\\\\d+$'", str(c1))
        self.assertEqual('"like"."foo" RLIKE \'\\\\d+$\'', str(c2))

    def test_glob_single_chars_and_various_chars(self):
        c1 = Field("foo").glob("a_b*")
        c2 = Field("foo", table=self.t).glob("a_b*")

        self.assertEqual("\"foo\" GLOB 'a_b*'", str(c1))
        self.assertEqual('"like"."foo" GLOB \'a_b*\'', str(c2))


class ExistsCriterionTests(unittest.TestCase):
    t2 = Table("abc", alias="t2")
    q2 = QueryBuilder().from_(t2).select(t2.field2)

    def test_exists(self):
        t1 = Table("def", alias="t1")
        q1 = QueryBuilder().from_(t1).where(ExistsCriterion(self.q2)).select(t1.field1)

        self.assertEqual(
            'SELECT "t1"."field1" FROM "def" "t1" WHERE EXISTS (SELECT "t2"."field2" FROM "abc" "t2")',
            str(q1),
        )

    def test_not_exists(self):
        t1 = Table("def", alias="t1")
        q1 = QueryBuilder().from_(t1).where(ExistsCriterion(self.q2).negate()).select(t1.field1)

        self.assertEqual(
            'SELECT "t1"."field1" FROM "def" "t1" WHERE NOT EXISTS (SELECT "t2"."field2" FROM "abc" "t2")',
            str(q1),
        )


class ComplexCriterionTests(unittest.TestCase):
    table_abc, table_efg = Table("abc", alias="cx0"), Table("efg", alias="cx1")

    def test_and(self):
        c1 = (Field("foo") == 1) & (Field("bar") == 2)
        c2 = (Field("foo", table=self.table_abc) == 1) & (Field("bar", table=self.table_efg) == 2)

        self.assertEqual('"foo"=1 AND "bar"=2', str(c1))
        self.assertEqual('"cx0"."foo"=1 AND "cx1"."bar"=2', str(c2))

    def test_or(self):
        c1 = (Field("foo") == 1) | (Field("bar") == 2)
        c2 = (Field("foo", table=self.table_abc) == 1) | (Field("bar", table=self.table_efg) == 2)

        self.assertEqual('"foo"=1 OR "bar"=2', str(c1))
        self.assertEqual('"cx0"."foo"=1 OR "cx1"."bar"=2', str(c2))

    def test_xor(self):
        c1 = (Field("foo") == 1) ^ (Field("bar") == 2)
        c2 = (Field("foo", table=self.table_abc) == 1) ^ (Field("bar", table=self.table_efg) == 2)

        self.assertEqual('"foo"=1 XOR "bar"=2', str(c1))
        self.assertEqual('"cx0"."foo"=1 XOR "cx1"."bar"=2', str(c2))

    def test_function_and(self):
        c1 = fn.IsNull(Field("foo")) & (Field("bar") == 2)

        self.assertEqual('ISNULL("foo") AND "bar"=2', str(c1))

    def test_function_or(self):
        c1 = fn.IsNull(Field("foo")) | (Field("bar") == 2)

        self.assertEqual('ISNULL("foo") OR "bar"=2', str(c1))

    def test_function_xor(self):
        c1 = fn.IsNull(Field("foo")) ^ (Field("bar") == 2)

        self.assertEqual('ISNULL("foo") XOR "bar"=2', str(c1))

    def test__nested__and(self):
        c = (Field("foo") == 1) & (Field("bar") == 2) & (Field("buz") == 3)

        self.assertEqual('"foo"=1 AND "bar"=2 AND "buz"=3', str(c))

    def test__nested__or(self):
        c = (Field("foo") == 1) | (Field("bar") == 2) | (Field("buz") == 3)

        self.assertEqual('"foo"=1 OR "bar"=2 OR "buz"=3', str(c))

    def test__nested__mixed(self):
        c = ((Field("foo") == 1) & (Field("bar") == 2)) | (Field("buz") == 3)

        self.assertEqual('("foo"=1 AND "bar"=2) OR "buz"=3', str(c))

    def test__between_and_isin(self):
        c = Field("foo").isin([1, 2, 3]) & Field("bar").between(0, 1)

        self.assertEqual('"foo" IN (1,2,3) AND "bar" BETWEEN 0 AND 1', str(c))

    def test__between_and_field(self):
        c1 = Field("foo").between(0, 1)
        c2 = Field("bool_field")

        self.assertEqual('"foo" BETWEEN 0 AND 1 AND "bool_field"', str(c1 & c2))
        self.assertEqual('"bool_field" AND "foo" BETWEEN 0 AND 1', str(c2 & c1))


class FieldsAsCriterionTests(unittest.TestCase):
    def test__field_and_field(self):
        c1 = Field("a")
        c2 = Field("b")

        self.assertEqual('"a" AND "b"', str(c1 & c2))

    def test__field_or_field(self):
        c1 = Field("a")
        c2 = Field("b")

        self.assertEqual('"a" OR "b"', str(c1 | c2))

    def test__field_xor_field(self):
        c1 = Field("a")
        c2 = Field("b")

        self.assertEqual('"a" XOR "b"', str(c1 ^ c2))


class CriterionOperationsTests(unittest.TestCase):
    table_abc, table_efg = Table("abc", alias="cx0"), Table("efg", alias="cx1")

    def test_field_replace_table(self):
        f = self.table_abc.foo.replace_table(self.table_abc, self.table_efg)

        self.assertEqual('"cx1"."foo"', str(f))

    def test_arithmeticfunction_replace_table(self):
        f = (self.table_abc.foo + self.table_abc.bar).replace_table(self.table_abc, self.table_efg)

        self.assertEqual('"cx1"."foo"+"cx1"."bar"', str(f))

    def test_criterion_replace_table(self):
        f = (self.table_abc.foo < self.table_abc.bar).replace_table(self.table_abc, self.table_efg)

        self.assertEqual('"cx1"."foo"<"cx1"."bar"', str(f))

    def test_complexcriterion_replace_table(self):
        f = (self.table_abc.foo < self.table_abc.bar) & (self.table_abc.fiz > self.table_abc.buz)
        f = f.replace_table(self.table_abc, self.table_efg)

        self.assertEqual('"cx1"."foo"<"cx1"."bar" AND "cx1"."fiz">"cx1"."buz"', str(f))

    def test_function_with_only_fields_replace_table(self):
        f = fn.Sum(self.table_abc.foo).replace_table(self.table_abc, self.table_efg)

        self.assertEqual('SUM("cx1"."foo")', str(f))

    def test_function_with_values_and_fields_replace_table(self):
        f = Mod(self.table_abc.foo, 2).replace_table(self.table_abc, self.table_efg)

        self.assertEqual('MOD("cx1"."foo",2)', str(f))

    def test_betweencriterion_replace_table(self):
        f = self.table_abc.foo[0:1].replace_table(self.table_abc, self.table_efg)

        self.assertEqual('"cx1"."foo" BETWEEN 0 AND 1', str(f))

    def test_nullcriterion_replace_table(self):
        f = self.table_abc.foo.isnull().replace_table(self.table_abc, self.table_efg)

        self.assertEqual('"cx1"."foo" IS NULL', str(f))


class AnyTests(unittest.TestCase):
    def test_zero_args_returns_empty_criterion(self):
        crit = Criterion.any()
        self.assertIsInstance(crit, EmptyCriterion)

    def test_single_arg_returns_self(self):
        f = Field("a")
        crit = Criterion.any([f])
        self.assertEqual(str(f), str(crit))

    def test_multiple_args_returned_in_chain_of_ors(self):
        crit = Criterion.any([Field("a"), Field("b"), Field("c"), Field("d")])
        self.assertEqual(str(crit), '"a" OR "b" OR "c" OR "d"')

    def test_with_generator(self):
        crit = Criterion.any(Field(letter) for letter in "abcd")
        self.assertEqual(str(crit), '"a" OR "b" OR "c" OR "d"')


class EmptyCriterionTests(unittest.TestCase):
    def test_fields_(self):
        empty_criterion = EmptyCriterion()

        self.assertEqual(len(empty_criterion.fields_()), 0)

    def test_empty_criterion_on_the_left(self):
        t = Table("test1")
        q = EmptyCriterion() & t.f1 == "v1"

        self.assertEqual(
            '"test1"."f1"=\'v1\'',
            q.get_sql(with_namespace=True),
        )

    def test_empty_criterion_on_the_right(self):
        t = Table("test1")
        q = (t.f1 == "v1") & EmptyCriterion()

        self.assertEqual(
            '"test1"."f1"=\'v1\'',
            q.get_sql(with_namespace=True),
        )

    def test_invertion_of_the_empty_criterion(self):
        t = Table("test1")
        q = (t.f1 == "v1") & ~EmptyCriterion()

        self.assertEqual(
            '"test1"."f1"=\'v1\'',
            q.get_sql(with_namespace=True),
        )

    def test_different_operations_with_empty_criterion(self):
        t = Table("test1")
        q1 = (t.f1 == "v1") & EmptyCriterion()
        q2 = (t.f1 == "v1") | EmptyCriterion()
        q3 = (t.f1 == "v1") ^ EmptyCriterion()

        expected_sql = '"test1"."f1"=\'v1\''
        self.assertEqual(
            expected_sql,
            q1.get_sql(with_namespace=True),
        )
        self.assertEqual(
            expected_sql,
            q2.get_sql(with_namespace=True),
        )
        self.assertEqual(
            expected_sql,
            q3.get_sql(with_namespace=True),
        )

    def test_more_than_one_empty_criterions(self):
        t = Table("test1")
        q = EmptyCriterion() & ~(EmptyCriterion() | EmptyCriterion()) & (t.f1 == "v1") & ~EmptyCriterion()

        self.assertEqual(
            '"test1"."f1"=\'v1\'',
            q.get_sql(with_namespace=True),
        )


class AllTests(unittest.TestCase):
    def test_zero_args_returns_empty_criterion(self):
        crit = Criterion.all()
        self.assertIsInstance(crit, EmptyCriterion)

    def test_single_arg_returns_self(self):
        f = Field("a")
        crit = Criterion.all([f])
        self.assertEqual(str(f), str(crit))

    def test_multiple_args_returned_in_chain_of_ors(self):
        crit = Criterion.all([Field("a"), Field("b"), Field("c"), Field("d")])
        self.assertEqual(str(crit), '"a" AND "b" AND "c" AND "d"')

    def test_with_generator(self):
        crit = Criterion.all(Field(letter) for letter in "abcd")
        self.assertEqual(str(crit), '"a" AND "b" AND "c" AND "d"')



import unittest

from parameterized import parameterized

from pypika import Field
from pypika.clickhouse.nullable_arg import IfNull
from pypika.clickhouse.type_conversion import ToFixedString


class TestSearchString(unittest.TestCase):
    @parameterized.expand(
        [
            (
                IfNull(Field("name"), Field("login")),
                "ifNull(name,login)",
            ),
            (
                IfNull(Field("builder"), ToFixedString("pypika", 100)),
                "ifNull(builder,toFixedString('pypika',100))",
            ),
        ],
    )
    def test_get_sql(self, func, expected):
        self.assertEqual(func.get_sql(), expected)



import unittest
from datetime import date

from pypika import (
    FormatParameter,
    NamedParameter,
    NumericParameter,
    Parameter,
    PyformatParameter,
    QmarkParameter,
    Query,
    Tables,
)
from pypika.terms import ListParameter, ParameterValueWrapper


class ParametrizedTests(unittest.TestCase):
    table_abc, table_efg = Tables("abc", "efg")

    def test_param_insert(self):
        q = Query.into(self.table_abc).columns("a", "b", "c").insert(Parameter("?"), Parameter("?"), Parameter("?"))

        self.assertEqual('INSERT INTO "abc" ("a","b","c") VALUES (?,?,?)', q.get_sql())

    def test_param_select_join(self):
        q = (
            Query.from_(self.table_abc)
            .select("*")
            .where(self.table_abc.category == Parameter("%s"))
            .join(self.table_efg)
            .on(self.table_abc.id == self.table_efg.abc_id)
            .where(self.table_efg.date >= Parameter("%s"))
            .limit(10)
        )

        self.assertEqual(
            'SELECT * FROM "abc" JOIN "efg" ON "abc"."id"="efg"."abc_id" '
            'WHERE "abc"."category"=%s AND "efg"."date">=%s LIMIT 10',
            q.get_sql(),
        )

    def test_param_select_subquery(self):
        q = (
            Query.from_(self.table_abc)
            .select("*")
            .where(self.table_abc.category == Parameter("&1"))
            .where(
                self.table_abc.id.isin(
                    Query.from_(self.table_efg)
                    .select(self.table_efg.abc_id)
                    .where(self.table_efg.date >= Parameter("&2")),
                ),
            )
            .limit(10)
        )

        self.assertEqual(
            'SELECT * FROM "abc" WHERE "category"=&1 AND "id" '
            'IN (SELECT "abc_id" FROM "efg" WHERE "date">=&2) LIMIT 10',
            q.get_sql(),
        )

    def test_join(self):
        subquery = (
            Query.from_(self.table_efg)
            .select(self.table_efg.fiz, self.table_efg.buz)
            .where(self.table_efg.buz == Parameter(":buz"))
        )

        q = (
            Query.from_(self.table_abc)
            .join(subquery)
            .on(self.table_abc.bar == subquery.buz)
            .select(self.table_abc.foo, subquery.fiz)
            .where(self.table_abc.bar == Parameter(":bar"))
        )

        self.assertEqual(
            'SELECT "abc"."foo","sq0"."fiz" FROM "abc" JOIN (SELECT "fiz","buz" FROM "efg" WHERE "buz"=:buz)'
            ' "sq0" ON "abc"."bar"="sq0"."buz" WHERE "abc"."bar"=:bar',
            q.get_sql(),
        )

    def test_qmark_parameter(self):
        self.assertEqual('?', QmarkParameter().get_sql())

    def test_numeric_parameter(self):
        self.assertEqual(':14', NumericParameter('14').get_sql())
        self.assertEqual(':15', NumericParameter(15).get_sql())

    def test_named_parameter(self):
        self.assertEqual(':buz', NamedParameter('buz').get_sql())

    def test_format_parameter(self):
        self.assertEqual('%s', FormatParameter().get_sql())

    def test_pyformat_parameter(self):
        self.assertEqual('%(buz)s', PyformatParameter('buz').get_sql())


class ParametrizedTestsWithValues(unittest.TestCase):
    table_abc, table_efg = Tables("abc", "efg")

    def test_param_insert(self):
        q = Query.into(self.table_abc).columns("a", "b", "c").insert(1, 2.2, 'foo')

        parameter = QmarkParameter()
        sql = q.get_sql(parameter=parameter)
        self.assertEqual('INSERT INTO "abc" ("a","b","c") VALUES (?,?,?)', sql)
        self.assertEqual([1, 2.2, 'foo'], parameter.get_parameters())

    def test_param_select_join(self):
        q = (
            Query.from_(self.table_abc)
            .select("*")
            .where(self.table_abc.category == 'foobar')
            .join(self.table_efg)
            .on(self.table_abc.id == self.table_efg.abc_id)
            .where(self.table_efg.date >= date(2024, 2, 22))
            .limit(10)
        )

        parameter = FormatParameter()
        sql = q.get_sql(parameter=parameter)
        self.assertEqual(
            'SELECT * FROM "abc" JOIN "efg" ON "abc"."id"="efg"."abc_id" '
            'WHERE "abc"."category"=%s AND "efg"."date">=%s LIMIT 10',
            sql,
        )
        self.assertEqual(['foobar', '2024-02-22'], parameter.get_parameters())

    def test_param_select_subquery(self):
        q = (
            Query.from_(self.table_abc)
            .select("*")
            .where(self.table_abc.category == 'foobar')
            .where(
                self.table_abc.id.isin(
                    Query.from_(self.table_efg)
                    .select(self.table_efg.abc_id)
                    .where(self.table_efg.date >= date(2024, 2, 22)),
                ),
            )
            .limit(10)
        )

        parameter = ListParameter(placeholder=lambda idx: f'&{idx + 1}')
        sql = q.get_sql(parameter=parameter)
        self.assertEqual(
            'SELECT * FROM "abc" WHERE "category"=&1 AND "id" '
            'IN (SELECT "abc_id" FROM "efg" WHERE "date">=&2) LIMIT 10',
            sql,
        )
        self.assertEqual(['foobar', '2024-02-22'], parameter.get_parameters())

    def test_join(self):
        subquery = (
            Query.from_(self.table_efg)
            .select(self.table_efg.fiz, self.table_efg.buz)
            .where(self.table_efg.buz == 'buz')
        )

        q = (
            Query.from_(self.table_abc)
            .join(subquery)
            .on(self.table_abc.bar == subquery.buz)
            .select(self.table_abc.foo, subquery.fiz)
            .where(self.table_abc.bar == 'bar')
        )

        parameter = NamedParameter()
        sql = q.get_sql(parameter=parameter)
        self.assertEqual(
            'SELECT "abc"."foo","sq0"."fiz" FROM "abc" JOIN (SELECT "fiz","buz" FROM "efg" WHERE "buz"=:param1)'
            ' "sq0" ON "abc"."bar"="sq0"."buz" WHERE "abc"."bar"=:param2',
            sql,
        )
        self.assertEqual({'param1': 'buz', 'param2': 'bar'}, parameter.get_parameters())

    def test_join_with_parameter_value_wrapper(self):
        subquery = (
            Query.from_(self.table_efg)
            .select(self.table_efg.fiz, self.table_efg.buz)
            .where(self.table_efg.buz == ParameterValueWrapper(Parameter(':buz'), 'buz'))
        )

        q = (
            Query.from_(self.table_abc)
            .join(subquery)
            .on(self.table_abc.bar == subquery.buz)
            .select(self.table_abc.foo, subquery.fiz)
            .where(self.table_abc.bar == ParameterValueWrapper(NamedParameter('bar'), 'bar'))
        )

        parameter = NamedParameter()
        sql = q.get_sql(parameter=parameter)
        self.assertEqual(
            'SELECT "abc"."foo","sq0"."fiz" FROM "abc" JOIN (SELECT "fiz","buz" FROM "efg" WHERE "buz"=:buz)'
            ' "sq0" ON "abc"."bar"="sq0"."buz" WHERE "abc"."bar"=:bar',
            sql,
        )
        self.assertEqual({':buz': 'buz', 'bar': 'bar'}, parameter.get_parameters())

    def test_pyformat_parameter(self):
        q = Query.into(self.table_abc).columns("a", "b", "c").insert(1, 2.2, 'foo')

        parameter = PyformatParameter()
        sql = q.get_sql(parameter=parameter)
        self.assertEqual('INSERT INTO "abc" ("a","b","c") VALUES (%(param1)s,%(param2)s,%(param3)s)', sql)
        self.assertEqual({"param1": 1, "param2": 2.2, "param3": "foo"}, parameter.get_parameters())



import unittest

from pypika import (
    Array,
    Bracket,
    PostgreSQLQuery,
    Query,
    Table,
    Tables,
    Tuple,
)
from pypika.functions import Coalesce, NullIf, Sum
from pypika.terms import Field


class TupleTests(unittest.TestCase):
    table_abc, table_efg = Tables("abc", "efg")

    def test_tuple_equality_tuple_on_both(self):
        q = (
            Query.from_(self.table_abc)
            .select(self.table_abc.foo, self.table_abc.bar)
            .where(Tuple(self.table_abc.foo, self.table_abc.bar) == Tuple(1, 2))
        )

        self.assertEqual('SELECT "foo","bar" FROM "abc" WHERE ("foo","bar")=(1,2)', str(q))

    def test_tuple_equality_tuple_on_left(self):
        q = (
            Query.from_(self.table_abc)
            .select(self.table_abc.foo, self.table_abc.bar)
            .where(Tuple(self.table_abc.foo, self.table_abc.bar) == (1, 2))
        )

        self.assertEqual('SELECT "foo","bar" FROM "abc" WHERE ("foo","bar")=(1,2)', str(q))

    def test_tuple_equality_tuple_on_right(self):
        q = (
            Query.from_(self.table_abc)
            .select(self.table_abc.foo, self.table_abc.bar)
            .where((self.table_abc.foo, self.table_abc.bar) == Tuple(1, 2))
        )

        # Order is reversed due to lack of right equals method
        self.assertEqual('SELECT "foo","bar" FROM "abc" WHERE (1,2)=("foo","bar")', str(q))

    def test_tuple_in_using_python_tuples(self):
        q = (
            Query.from_(self.table_abc)
            .select(self.table_abc.foo, self.table_abc.bar)
            .where(Tuple(self.table_abc.foo, self.table_abc.bar).isin([(1, 1), (2, 2), (3, 3)]))
        )

        self.assertEqual(
            'SELECT "foo","bar" FROM "abc" WHERE ("foo","bar") IN ((1,1),(2,2),(3,3))',
            str(q),
        )

    def test_tuple_in_using_pypika_tuples(self):
        q = (
            Query.from_(self.table_abc)
            .select(self.table_abc.foo, self.table_abc.bar)
            .where(Tuple(self.table_abc.foo, self.table_abc.bar).isin([Tuple(1, 1), Tuple(2, 2), Tuple(3, 3)]))
        )

        self.assertEqual(
            'SELECT "foo","bar" FROM "abc" WHERE ("foo","bar") IN ((1,1),(2,2),(3,3))',
            str(q),
        )

    def test_tuple_in_using_mixed_tuples(self):
        q = (
            Query.from_(self.table_abc)
            .select(self.table_abc.foo, self.table_abc.bar)
            .where(Tuple(self.table_abc.foo, self.table_abc.bar).isin([(1, 1), Tuple(2, 2), (3, 3)]))
        )

        self.assertEqual(
            'SELECT "foo","bar" FROM "abc" WHERE ("foo","bar") IN ((1,1),(2,2),(3,3))',
            str(q),
        )

    def test_tuples_in_join(self):
        query = (
            Query.from_(self.table_abc)
            .join(self.table_efg)
            .on(self.table_abc.foo == self.table_efg.bar)
            .select("*")
            .where(Tuple(self.table_abc.foo, self.table_efg.bar).isin([(1, 1), Tuple(2, 2), (3, 3)]))
        )

        self.assertEqual(
            'SELECT * FROM "abc" JOIN "efg" ON "abc"."foo"="efg"."bar" '
            'WHERE ("abc"."foo","efg"."bar") IN ((1,1),(2,2),(3,3))',
            str(query),
        )

    def test_render_alias_in_array_sql(self):
        tb = Table("tb")

        q = Query.from_(tb).select(Tuple(tb.col).as_("different_name"))
        self.assertEqual(str(q), 'SELECT ("col") "different_name" FROM "tb"')

    def test_tuple_is_aggregate(self):
        with self.subTest('None if single argument returns None for is_aggregate'):
            self.assertEqual(None, Tuple(0).is_aggregate)
            self.assertEqual(None, Tuple(Coalesce('col')).is_aggregate)

        with self.subTest('None if multiple arguments all return None for is_aggregate'):
            self.assertEqual(None, Tuple(0, 'a').is_aggregate)
            self.assertEqual(None, Tuple(Coalesce('col'), NullIf('col2', 0)).is_aggregate)

        with self.subTest('True if single argument returns True for is_aggregate'):
            self.assertEqual(True, Tuple(Sum('col')).is_aggregate)

        with self.subTest('True if multiple arguments return True for is_aggregate'):
            self.assertEqual(True, Tuple(Sum('col'), Sum('col2')).is_aggregate)

        with self.subTest('True when mix of arguments returning None and True for is_aggregate'):
            self.assertEqual(True, Tuple(Coalesce('col'), Coalesce('col2', 0), Sum('col3')).is_aggregate)

        with self.subTest('False when one of the arguments returns False for is_aggregate'):
            self.assertEqual(False, Tuple(Field('col1'), Sum('col2')).is_aggregate)


class ArrayTests(unittest.TestCase):
    table_abc, table_efg = Tables("abc", "efg")

    def test_array_general(self):
        query = Query.from_(self.table_abc).select(Array(1, "a", ["b", 2, 3]))

        self.assertEqual("SELECT [1,'a',['b',2,3]] FROM \"abc\"", str(query))

    def test_empty_psql_array(self):
        query = PostgreSQLQuery.from_(self.table_abc).select(Array())

        self.assertEqual("SELECT '{}' FROM \"abc\"", str(query))

    def test_psql_array_general(self):
        query = PostgreSQLQuery.from_(self.table_abc).select(Array(1, Array(2, 2, 2), 3))

        self.assertEqual("SELECT ARRAY[1,ARRAY[2,2,2],3] FROM \"abc\"", str(query))

    def test_render_alias_in_array_sql(self):
        tb = Table("tb")

        q = Query.from_(tb).select(Array(tb.col).as_("different_name"))
        self.assertEqual(str(q), 'SELECT ["col"] "different_name" FROM "tb"')


class BracketTests(unittest.TestCase):
    table_abc, table_efg = Tables("abc", "efg")

    def test_arithmetic_with_brackets(self):
        q = Query.from_(self.table_abc).select(Bracket(self.table_abc.foo / 2) / 2)

        self.assertEqual('SELECT ("foo"/2)/2 FROM "abc"', str(q))

    def test_arithmetic_with_brackets_and_alias(self):
        q = Query.from_(self.table_abc).select(Bracket(self.table_abc.foo / 2).as_("alias"))

        self.assertEqual('SELECT ("foo"/2) "alias" FROM "abc"', str(q))



import unittest
from collections import OrderedDict

from pypika import (
    JSON,
    Array,
    Field,
    QueryException,
    Table,
)
from pypika.dialects import PostgreSQLQuery


class InsertTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_array_keyword(self):
        q = PostgreSQLQuery.into(self.table_abc).insert(1, [1, "a", True])

        self.assertEqual("INSERT INTO \"abc\" VALUES (1,ARRAY[1,'a',true])", str(q))


class JSONObjectTests(unittest.TestCase):
    def test_alias_set_correctly(self):
        table = Table('jsonb_table')
        q = PostgreSQLQuery.from_('abc').select(table.value.get_text_value('a').as_('name'))

        self.assertEqual('''SELECT "value"->>'a' "name" FROM "abc"''', str(q))

    def test_json_value_from_dict(self):
        q = PostgreSQLQuery.select(JSON({"a": "foo"}))

        self.assertEqual('SELECT \'{"a":"foo"}\'', str(q))

    def test_json_value_from_array_num(self):
        q = PostgreSQLQuery.select(JSON([1, 2, 3]))

        self.assertEqual("SELECT '[1,2,3]'", str(q))

    def test_json_value_from_array_str(self):
        q = PostgreSQLQuery.select(JSON(["a", "b", "c"]))

        self.assertEqual('SELECT \'["a","b","c"]\'', str(q))

    def test_json_value_from_dict_recursive(self):
        q = PostgreSQLQuery.select(JSON({"a": "z", "b": {"c": "foo"}, "d": 1}))

        # gotta split this one up to avoid the indeterminate order
        sql = str(q)
        start, end = 9, -2
        self.assertEqual("SELECT '{}'", sql[:start] + sql[end:])

        members_set = set(sql[start:end].split(","))
        self.assertSetEqual({'"a":"z"', '"b":{"c":"foo"}', '"d":1'}, members_set)


class JSONOperatorsTests(unittest.TestCase):
    # reference https://www.postgresql.org/docs/9.5/functions-json.html
    table_abc = Table("abc")

    def test_get_json_value_by_key(self):
        q = PostgreSQLQuery.from_(self.table_abc).select("*").where(self.table_abc.json.get_json_value("dates"))

        self.assertEqual('SELECT * FROM "abc" WHERE "json"->\'dates\'', str(q))

    def test_get_json_value_by_index(self):
        q = PostgreSQLQuery.from_(self.table_abc).select("*").where(self.table_abc.json.get_json_value(1))

        self.assertEqual('SELECT * FROM "abc" WHERE "json"->1', str(q))

    def test_get_text_value_by_key(self):
        q = PostgreSQLQuery.from_(self.table_abc).select("*").where(self.table_abc.json.get_text_value("dates"))

        self.assertEqual('SELECT * FROM "abc" WHERE "json"->>\'dates\'', str(q))

    def test_get_text_value_by_index(self):
        q = PostgreSQLQuery.from_(self.table_abc).select("*").where(self.table_abc.json.get_text_value(1))

        self.assertEqual('SELECT * FROM "abc" WHERE "json"->>1', str(q))

    def test_get_path_json_value(self):
        q = PostgreSQLQuery.from_(self.table_abc).select("*").where(self.table_abc.json.get_path_json_value("{a,b}"))

        self.assertEqual('SELECT * FROM "abc" WHERE "json"#>\'{a,b}\'', str(q))

    def test_get_path_text_value(self):
        q = PostgreSQLQuery.from_(self.table_abc).select("*").where(self.table_abc.json.get_path_text_value("{a,b}"))

        self.assertEqual('SELECT * FROM "abc" WHERE "json"#>>\'{a,b}\'', str(q))


class JSONBOperatorsTests(unittest.TestCase):
    # reference https://www.postgresql.org/docs/9.5/functions-json.html
    table_abc = Table("abc")

    def test_json_contains_for_json(self):
        q = PostgreSQLQuery.select(JSON({"a": 1, "b": 2}).contains({"a": 1}))

        # gotta split this one up to avoid the indeterminate order
        sql = str(q)
        start, end = 9, -13
        self.assertEqual("SELECT '{}'@>'{\"a\":1}'", sql[:start] + sql[end:])

        members_set = set(sql[start:end].split(","))
        self.assertSetEqual({'"a":1', '"b":2'}, members_set)

    def test_json_contains_for_field(self):
        q = (
            PostgreSQLQuery.from_(self.table_abc)
            .select("*")
            .where(self.table_abc.json.contains({"dates": "2018-07-10 - 2018-07-17"}))
        )

        self.assertEqual(
            "SELECT * FROM \"abc\" WHERE \"json\"@>'{\"dates\":\"2018-07-10 - 2018-07-17\"}'",
            str(q),
        )

    def test_json_contained_by_using_str_arg(self):
        q = (
            PostgreSQLQuery.from_(self.table_abc)
            .select("*")
            .where(
                self.table_abc.json.contained_by(
                    OrderedDict(
                        [
                            ("dates", "2018-07-10 - 2018-07-17"),
                            ("imported", "8"),
                        ],
                    ),
                ),
            )
        )
        self.assertEqual(
            'SELECT * FROM "abc" WHERE "json"<@\'{"dates":"2018-07-10 - 2018-07-17","imported":"8"}\'',
            str(q),
        )

    def test_json_contained_by_using_list_arg(self):
        q = (
            PostgreSQLQuery.from_(self.table_abc)
            .select("*")
            .where(self.table_abc.json.contained_by(["One", "Two", "Three"]))
        )

        self.assertEqual('SELECT * FROM "abc" WHERE "json"<@\'["One","Two","Three"]\'', str(q))

    def test_json_contained_by_with_complex_criterion(self):
        q = (
            PostgreSQLQuery.from_(self.table_abc)
            .select("*")
            .where(self.table_abc.json.contained_by(["One", "Two", "Three"]) & (self.table_abc.id == 26))
        )

        self.assertEqual(
            'SELECT * FROM "abc" WHERE "json"<@\'["One","Two","Three"]\' AND "id"=26',
            str(q),
        )

    def test_json_has_key(self):
        q = PostgreSQLQuery.from_(self.table_abc).select("*").where(self.table_abc.json.has_key("dates"))

        self.assertEqual('SELECT * FROM "abc" WHERE "json"?\'dates\'', str(q))

    def test_json_has_keys(self):
        q = PostgreSQLQuery.from_(self.table_abc).select("*").where(self.table_abc.json.has_keys(["dates", "imported"]))

        self.assertEqual("SELECT * FROM \"abc\" WHERE \"json\"?&ARRAY['dates','imported']", str(q))

    def test_json_has_any_keys(self):
        q = (
            PostgreSQLQuery.from_(self.table_abc)
            .select("*")
            .where(self.table_abc.json.has_any_keys(["dates", "imported"]))
        )

        self.assertEqual("SELECT * FROM \"abc\" WHERE \"json\"?|ARRAY['dates','imported']", str(q))

    def test_subnet_contains_inet(self):
        q = (
            PostgreSQLQuery.from_(self.table_abc)
            .select(self.table_abc.a.lshift(2))
            .where(self.table_abc.cidr >> "1.1.1.1")
        )

        self.assertEqual("SELECT \"a\"<<2 FROM \"abc\" WHERE \"cidr\">>'1.1.1.1'", str(q))


class DistinctOnTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_distinct_on(self):
        q = PostgreSQLQuery.from_(self.table_abc).distinct_on("lname", self.table_abc.fname).select("lname", "id")

        self.assertEqual('''SELECT DISTINCT ON("lname","fname") "lname","id" FROM "abc"''', str(q))


class ArrayTests(unittest.TestCase):
    def test_array_syntax(self):
        tb = Table("tb")
        q = PostgreSQLQuery.from_(tb).select(Array(1, "a", ["b", 2, 3]))

        self.assertEqual(str(q), "SELECT ARRAY[1,'a',ARRAY['b',2,3]] FROM \"tb\"")

    def test_render_alias_in_array_sql(self):
        tb = Table("tb")

        q = PostgreSQLQuery.from_(tb).select(Array(tb.col).as_("different_name"))
        self.assertEqual(str(q), 'SELECT ARRAY["col"] "different_name" FROM "tb"')


class ReturningClauseTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.table_abc = Table('abc')

    def test_returning_from_missing_table_raises_queryexception(self):
        field_from_diff_table = Field('xyz', table=Table('other'))

        with self.assertRaisesRegex(QueryException, "You can't return from other tables"):
            (
                PostgreSQLQuery.from_(self.table_abc)
                .where(self.table_abc.foo == self.table_abc.bar)
                .delete()
                .returning(field_from_diff_table)
            )

    def test_queryexception_if_returning_used_on_invalid_query(self):
        with self.assertRaisesRegex(QueryException, "Returning can't be used in this query"):
            PostgreSQLQuery.from_(self.table_abc).select('abc').returning('abc')

    def test_no_queryexception_if_returning_used_on_valid_query_type(self):
        # No exceptions for insert, update and delete queries
        with self.subTest('DELETE'):
            PostgreSQLQuery.from_(self.table_abc).where(self.table_abc.foo == self.table_abc.bar).delete().returning(
                "id",
            )
        with self.subTest('UPDATE'):
            PostgreSQLQuery.update(self.table_abc).where(self.table_abc.foo == 0).set("foo", "bar").returning("id")
        with self.subTest('INSERT'):
            PostgreSQLQuery.into(self.table_abc).insert('abc').returning('abc')

    def test_return_field_from_join_table(self):
        new_table = Table('xyz')
        q = (
            PostgreSQLQuery.update(self.table_abc)
            .join(new_table)
            .on(new_table.id == self.table_abc.xyz)
            .where(self.table_abc.foo == 0)
            .set("foo", "bar")
            .returning(new_table.a)
        )

        self.assertEqual(
            'UPDATE "abc" '
            'JOIN "xyz" ON "xyz"."id"="abc"."xyz" '
            'SET "foo"=\'bar\' '
            'WHERE "abc"."foo"=0 '
            'RETURNING "xyz"."a"',
            str(q),
        )



import unittest

from pypika import (
    Case,
    Field,
    Table,
)
from pypika import (
    functions as fn,
)
from pypika.terms import (
    Negative,
    ValueWrapper,
)


class IsAggregateTests(unittest.TestCase):
    def test__field_is_not_aggregate(self):
        v = Field("foo")
        self.assertFalse(v.is_aggregate)

    def test__constant_is_aggregate_none(self):
        v = ValueWrapper(100)
        self.assertIsNone(v.is_aggregate)

    def test__constant_arithmetic_is_aggregate_none(self):
        v = ValueWrapper(100) + ValueWrapper(100)
        self.assertIsNone(v.is_aggregate)

    def test__field_arithmetic_is_not_aggregate(self):
        v = Field("foo") + Field("bar")
        self.assertFalse(v.is_aggregate)

    def test__field_arithmetic_constant_is_not_aggregate(self):
        v = Field("foo") + 1
        self.assertFalse(v.is_aggregate)

    def test__agg_func_is_aggregate(self):
        v = fn.Sum(Field("foo"))
        self.assertTrue(v.is_aggregate)

    def test__negative_agg_func_is_aggregate(self):
        v = Negative(fn.Sum(Field("foo")))
        self.assertTrue(v.is_aggregate)

    def test__agg_func_arithmetic_is_aggregate(self):
        v = fn.Sum(Field("foo")) / fn.Sum(Field("foo"))
        self.assertTrue(v.is_aggregate)

    def test__mixed_func_arithmetic_is_not_aggregate(self):
        v = Field("foo") / fn.Sum(Field("foo"))
        self.assertFalse(v.is_aggregate)

    def test__func_arithmetic_constant_is_not_aggregate(self):
        v = 1 / fn.Sum(Field("foo"))
        self.assertTrue(v.is_aggregate)

    def test__agg_case_criterion_is_aggregate(self):
        v = Case().when(fn.Sum(Field("foo")) > 666, 'More than 666').else_('Less than 666')

        self.assertTrue(v.is_aggregate)

    def test__agg_case_is_aggregate(self):
        v = (
            Case()
            .when(Field("foo") == 1, fn.Sum(Field("bar")))
            .when(Field("foo") == 2, fn.Sum(Field("fiz")))
            .else_(fn.Sum(Field("fiz")))
        )

        self.assertTrue(v.is_aggregate)

    def test__mixed_case_is_not_aggregate(self):
        v = Case().when(Field("foo") == 1, fn.Sum(Field("bar"))).when(Field("foo") == 2, Field("fiz"))

        self.assertFalse(v.is_aggregate)

    def test__case_mixed_else_is_not_aggregate(self):
        v = (
            Case()
            .when(Field("foo") == 1, fn.Sum(Field("bar")))
            .when(Field("foo") == 2, fn.Sum(Field("fiz")))
            .else_(Field("fiz"))
        )

        self.assertFalse(v.is_aggregate)

    def test__case_mixed_constant_is_not_aggregate(self):
        v = Case().when(Field("foo") == 1, fn.Sum(Field("bar"))).when(Field("foo") == 2, fn.Sum(Field("fiz"))).else_(1)

        self.assertTrue(v.is_aggregate)

    def test__case_with_field_is_not_aggregate(self):
        v = Case().when(Field("foo") == 1, 1).when(Field("foo") == 2, 2).else_(3)

        self.assertFalse(v.is_aggregate)

    def test__case_with_single_aggregate_field_in_one_criterion_is_aggregate(self):
        v = Case().when(Field("foo") == 1, 1).when(fn.Sum(Field("foo")) == 2, 2).else_(3)

        self.assertTrue(v.is_aggregate)

    def test__non_aggregate_function_with_aggregated_arg(self):
        t = Table("abc")
        expr = fn.Sqrt(fn.Sum(t.a))

        self.assertTrue(expr.is_aggregate)

    def test_complicated(self):
        t = Table("abc")
        is_placebo = t.campaign_extra_info == "placebo"

        pixel_mobile_search = Case().when(is_placebo, t.action_fb_pixel_search + t.action_fb_mobile_search)
        unique_impressions = Case().when(is_placebo, t.unique_impressions)

        v = fn.Sum(pixel_mobile_search) / fn.Sum(unique_impressions) - 1.96 * fn.Sqrt(
            1
            / fn.Sum(unique_impressions)
            * fn.Sum(pixel_mobile_search)
            / fn.Sum(unique_impressions)
            * (1 - fn.Sum(pixel_mobile_search) / fn.Sum(unique_impressions)),
        )

        self.assertTrue(v.is_aggregate)



import unittest

from pypika import SYSTEM_TIME, PostgreSQLQuery, Query, Table

__author__ = "Timothy Heys"
__email__ = "theys@kayak.com"

from pypika.terms import Star


class DeleteTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.table_abc = Table("abc")

    def test_omit_where(self):
        q = Query.from_("abc").delete()

        self.assertEqual('DELETE FROM "abc"', str(q))

    def test_omit_where__table_schema(self):
        q = Query.from_(Table("abc", "schema1")).delete()

        self.assertEqual('DELETE FROM "schema1"."abc"', str(q))

    def test_where_field_equals(self):
        q1 = Query.from_(self.table_abc).where(self.table_abc.foo == self.table_abc.bar).delete()
        q2 = Query.from_(self.table_abc).where(self.table_abc.foo.eq(self.table_abc.bar)).delete()

        self.assertEqual('DELETE FROM "abc" WHERE "foo"="bar"', str(q1))
        self.assertEqual('DELETE FROM "abc" WHERE "foo"="bar"', str(q2))

    def test_for_portion(self):
        with self.subTest("with system time"):
            q = Query.from_(self.table_abc.for_portion(SYSTEM_TIME.from_to('2020-01-01', '2020-02-01'))).delete()

            self.assertEqual(
                'DELETE FROM "abc" FOR PORTION OF SYSTEM_TIME FROM \'2020-01-01\' TO \'2020-02-01\'',
                str(q),
            )

        with self.subTest("with column"):
            q = Query.from_(
                self.table_abc.for_portion(self.table_abc.valid_period.from_to('2020-01-01', '2020-02-01')),
            ).delete()

            self.assertEqual(
                'DELETE FROM "abc" FOR PORTION OF "valid_period" FROM \'2020-01-01\' TO \'2020-02-01\'',
                str(q),
            )


class PostgresDeleteTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.table_abc = Table("abc")

    def test_delete_returning(self):
        q1 = (
            PostgreSQLQuery.from_(self.table_abc)
            .where(self.table_abc.foo == self.table_abc.bar)
            .delete()
            .returning(self.table_abc.id)
        )

        self.assertEqual('DELETE FROM "abc" WHERE "foo"="bar" RETURNING "id"', str(q1))

    def test_delete_returning_str(self):
        q1 = (
            PostgreSQLQuery.from_(self.table_abc)
            .where(self.table_abc.foo == self.table_abc.bar)
            .delete()
            .returning("id")
        )

        self.assertEqual('DELETE FROM "abc" WHERE "foo"="bar" RETURNING "id"', str(q1))

    def test_delete_returning_star(self):
        q1 = (
            PostgreSQLQuery.from_(self.table_abc)
            .where(self.table_abc.foo == self.table_abc.bar)
            .delete()
            .returning(Star())
        )

        self.assertEqual('DELETE FROM "abc" WHERE "foo"="bar" RETURNING *', str(q1))

    def test_delete_using(self):
        table_trash = Table('trash')
        q1 = (
            PostgreSQLQuery.from_(self.table_abc)
            .using(table_trash)
            .where(self.table_abc.id == table_trash.abc_id)
            .delete()
        )

        self.assertEqual('DELETE FROM "abc" USING "trash" WHERE "abc"."id"="trash"."abc_id"', str(q1))



import unittest
import uuid

from pypika.terms import ValueWrapper


class StringTests(unittest.TestCase):
    def test_inline_string_concatentation(self):
        self.assertEqual("'it''s'", ValueWrapper("it's").get_sql())


class UuidTests(unittest.TestCase):
    def test_uuid_string_generation(self):
        id = uuid.uuid4()
        self.assertEqual(f"'{id}'", ValueWrapper(id).get_sql())



import unittest

from pypika import (
    Case,
    Field,
    Table,
)
from pypika.terms import Star


class TablesTests(unittest.TestCase):
    def test__criterion_replace_table_with_value(self):
        table = Table("a")

        c = (Field("foo") > 1).replace_table(None, table)
        self.assertEqual(c.left, table)
        self.assertEqual(c.tables_, {table})

    def test__case_tables(self):
        table = Table("a")

        c = Case().when(table.a == 1, 2 * table.a)
        self.assertIsInstance(c.tables_, set)
        self.assertSetEqual(c.tables_, {table})

    def test__star_tables(self):
        star = Star()

        self.assertEqual(star.tables_, set())

    def test__table_star_tables(self):
        table = Table("a")
        star = Star(table=table)

        self.assertEqual(star.tables_, {table})



import unittest

from pypika import Columns, Table, Tables, VerticaQuery


class VerticaQueryTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_select_query_with_hint(self):
        q = VerticaQuery.from_("abc").select("*").hint("test_hint")

        self.assertEqual('SELECT /*+label(test_hint)*/ * FROM "abc"', str(q))

    def test_insert_query_with_hint(self):
        query = VerticaQuery.into(self.table_abc).insert(1).hint("test_hint")

        self.assertEqual('INSERT /*+label(test_hint)*/ INTO "abc" VALUES (1)', str(query))

    def test_update_query_with_hint(self):
        q = VerticaQuery.update(self.table_abc).set("foo", "bar").hint("test_hint")

        self.assertEqual('UPDATE /*+label(test_hint)*/ "abc" SET "foo"=\'bar\'', str(q))

    def test_delete_query_with_hint(self):
        q = VerticaQuery.from_("abc").delete().hint("test_hint")

        self.assertEqual('DELETE /*+label(test_hint)*/ FROM "abc"', str(q))


class CopyCSVTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_copy_from_file(self):
        with self.subTest("with table as string"):
            q = VerticaQuery.from_file("/path/to/file").copy_("abc")

            self.assertEqual(
                "COPY \"abc\" FROM LOCAL '/path/to/file' PARSER fcsvparser(header=false)",
                str(q),
            )

        with self.subTest("with table"):
            q = VerticaQuery.from_file("/path/to/file").copy_(self.table_abc)

            self.assertEqual(
                "COPY \"abc\" FROM LOCAL '/path/to/file' PARSER fcsvparser(header=false)",
                str(q),
            )


class CreateTemporaryTableTests(unittest.TestCase):
    new_table, existing_table = Tables("abc", "efg")
    foo, bar = Columns(("a", "INT"), ("b", "VARCHAR(100)"))
    select = VerticaQuery.from_(existing_table).select(existing_table.foo, existing_table.bar)

    def test_create_local_temporary_table(self):
        with self.subTest("with columns"):
            q = VerticaQuery.create_table(self.new_table).temporary().local().columns(self.foo, self.bar)

            self.assertEqual('CREATE LOCAL TEMPORARY TABLE "abc" ("a" INT,"b" VARCHAR(100))', str(q))

        with self.subTest("with select"):
            q = VerticaQuery.create_table(self.new_table).temporary().local().as_select(self.select)

            self.assertEqual(
                'CREATE LOCAL TEMPORARY TABLE "abc" AS (SELECT "foo","bar" FROM "efg")',
                str(q),
            )

    def test_create_local_table_without_temporary_raises_error(self):
        with self.assertRaises(AttributeError):
            VerticaQuery.create_table(self.new_table).local()

    def test_create_temporary_table_preserve_rows(self):
        with self.subTest("with columns"):
            q = VerticaQuery.create_table(self.new_table).temporary().preserve_rows().columns(self.foo, self.bar)

            self.assertEqual(
                'CREATE TEMPORARY TABLE "abc" ("a" INT,"b" VARCHAR(100)) ON COMMIT PRESERVE ROWS',
                str(q),
            )

        with self.subTest("with select"):
            q = VerticaQuery.create_table(self.new_table).temporary().preserve_rows().as_select(self.select)

            self.assertEqual(
                'CREATE TEMPORARY TABLE "abc" ON COMMIT PRESERVE ROWS AS (SELECT "foo","bar" FROM "efg")',
                str(q),
            )

    def test_create_table_preserve_rows_without_temporary_raises_error(self):
        with self.assertRaises(AttributeError):
            VerticaQuery.create_table(self.new_table).preserve_rows()



import unittest

from parameterized import parameterized

from pypika import Field
from pypika.clickhouse.condition import If, MultiIf
from pypika.clickhouse.type_conversion import ToFixedString


class TestIfCondition(unittest.TestCase):
    @parameterized.expand(
        [
            (
                If(
                    Field("filmmaker").isnull(),
                    ToFixedString("Tarantino", 20),
                    Field("filmmaker"),
                ),
                "if(filmmaker IS NULL,toFixedString('Tarantino',20),filmmaker)",
            ),
            (
                If(
                    Field("created") == Field("updated"),
                    ToFixedString("yes", 3),
                    ToFixedString("no", 3),
                ),
                "if(created=updated,toFixedString('yes',3),toFixedString('no',3))",
            ),
        ],
    )
    def test_get_sql(self, func, expected):
        self.assertEqual(func.get_sql(), expected)


class TestMultiIfCondition(unittest.TestCase):
    @parameterized.expand(
        [
            (
                MultiIf(
                    Field("filmmaker").isnull(),
                    ToFixedString("Tarantino", 20),
                    Field("filmmaker") == ToFixedString("undefined", 20),
                    ToFixedString("Tarantino", 20),
                    Field("filmmaker"),
                ),
                "multiIf(filmmaker IS NULL,"
                "toFixedString('Tarantino',20),"
                "filmmaker=toFixedString('undefined',20),"
                "toFixedString('Tarantino',20),filmmaker)",
            ),
            (
                MultiIf(
                    Field("color") == ToFixedString("black", 20),
                    ToFixedString("dark", 20),
                    Field("color") == ToFixedString("grey", 20),
                    ToFixedString("dark", 20),
                    Field("color") == ToFixedString("white", 20),
                    ToFixedString("light", 20),
                    ToFixedString("undefined", 20),
                ),
                "multiIf(color=toFixedString('black',20),"
                "toFixedString('dark',20),"
                "color=toFixedString('grey',20),"
                "toFixedString('dark',20),"
                "color=toFixedString('white',20),"
                "toFixedString('light',20),"
                "toFixedString('undefined',20))",
            ),
        ],
    )
    def test_get_sql(self, func, expected):
        self.assertEqual(func.get_sql(), expected)



# from pypika.terms import ValueWrapper, SystemTimeValue
import unittest

from pypika import SYSTEM_TIME, Database, Dialects, Query, Schema, SQLLiteQuery, Table, Tables

__author__ = "Timothy Heys"
__email__ = "theys@kayak.com"


class TableStructureTests(unittest.TestCase):
    def test_table_sql(self):
        table = Table("test_table")

        self.assertEqual('"test_table"', str(table))

    def test_table_with_alias(self):
        table = Table("test_table").as_("my_table")

        self.assertEqual('"test_table" "my_table"', table.get_sql(with_alias=True, quote_char='"'))

    def test_schema_table_attr(self):
        table = Schema("x_schema").test_table

        self.assertEqual('"x_schema"."test_table"', str(table))

    def test_table_with_schema_arg(self):
        table = Table("test_table", schema=Schema("x_schema"))

        self.assertEqual('"x_schema"."test_table"', str(table))

    def test_database_schema_table_attr(self):
        table = Database("x_db").x_schema.test_table

        self.assertEqual('"x_db"."x_schema"."test_table"', str(table))

    def test_table_with_schema_and_schema_parent_arg(self):
        table = Table("test_table", schema=Schema("x_schema", parent=Database("x_db")))

        self.assertEqual('"x_db"."x_schema"."test_table"', str(table))

    def test_table_for_system_time_sql(self):
        with self.subTest("with between criterion"):
            table = Table("test_table").for_(SYSTEM_TIME.between('2020-01-01', '2020-02-01'))

            self.assertEqual('"test_table" FOR SYSTEM_TIME BETWEEN \'2020-01-01\' AND \'2020-02-01\'', str(table))

        with self.subTest("with as of criterion"):
            table = Table("test_table").for_(SYSTEM_TIME.as_of('2020-01-01'))

            self.assertEqual('"test_table" FOR SYSTEM_TIME AS OF \'2020-01-01\'', str(table))

        with self.subTest("with from to criterion"):
            table = Table("test_table").for_(SYSTEM_TIME.from_to('2020-01-01', '2020-02-01'))

            self.assertEqual('"test_table" FOR SYSTEM_TIME FROM \'2020-01-01\' TO \'2020-02-01\'', str(table))

    def test_table_for_period_sql(self):
        with self.subTest("with between criterion"):
            table = Table("test_table")
            table = table.for_(table.valid_period.between('2020-01-01', '2020-02-01'))

            self.assertEqual('"test_table" FOR "valid_period" BETWEEN \'2020-01-01\' AND \'2020-02-01\'', str(table))

        with self.subTest("with as of criterion"):
            table = Table("test_table")
            table = table.for_(table.valid_period.as_of('2020-01-01'))

            self.assertEqual('"test_table" FOR "valid_period" AS OF \'2020-01-01\'', str(table))

        with self.subTest("with from to criterion"):
            table = Table("test_table")
            table = table.for_(table.valid_period.from_to('2020-01-01', '2020-02-01'))

            self.assertEqual('"test_table" FOR "valid_period" FROM \'2020-01-01\' TO \'2020-02-01\'', str(table))


class TableEqualityTests(unittest.TestCase):
    def test_tables_equal_by_name(self):
        t1 = Table("t")
        t2 = Table("t")

        self.assertEqual(t1, t2)

    def test_tables_equal_by_schema_and_name(self):
        t1 = Table("t", schema="a")
        t2 = Table("t", schema="a")

        self.assertEqual(t1, t2)

    def test_tables_equal_by_schema_and_name_using_schema(self):
        a = Schema("a")
        t1 = Table("t", schema=a)
        t2 = Table("t", schema=a)

        self.assertEqual(t1, t2)

    def test_tables_equal_by_schema_and_name_using_schema_with_parent(self):
        parent = Schema("parent")
        a = Schema("a", parent=parent)
        t1 = Table("t", schema=a)
        t2 = Table("t", schema=a)

        self.assertEqual(t1, t2)

    def test_tables_not_equal_by_schema_and_name_using_schema_with_different_parents(
        self,
    ):
        parent = Schema("parent")
        a = Schema("a", parent=parent)
        t1 = Table("t", schema=a)
        t2 = Table("t", schema=Schema("a"))

        self.assertNotEqual(t1, t2)

    def test_tables_not_equal_with_different_schemas(self):
        t1 = Table("t", schema="a")
        t2 = Table("t", schema="b")

        self.assertNotEqual(t1, t2)

    def test_tables_not_equal_with_different_names(self):
        t1 = Table("t", schema="a")
        t2 = Table("q", schema="a")

        self.assertNotEqual(t1, t2)

    def test_many_tables_with_alias(self):
        tables_data = [("table1", "t1"), ("table2", "t2"), ("table3", "t3")]
        tables = Tables(*tables_data)
        for el in tables:
            self.assertIsNotNone(el.alias)

    def test_many_tables_without_alias(self):
        tables_data = ["table1", "table2", "table3"]
        tables = Tables(*tables_data)
        for el in tables:
            self.assertIsNone(el.alias)

    def test_many_tables_with_or_not_alias(self):
        tables_data = [("table1", "t1"), ("table2"), "table3"]
        tables = Tables(*tables_data)
        for i in range(len(tables)):
            if isinstance(tables_data[i], tuple):
                self.assertIsNotNone(tables[i].alias)
            else:
                self.assertIsNone(tables[i].alias)


class TableDialectTests(unittest.TestCase):
    def test_table_with_default_query_cls(self):
        table = Table("abc")
        q = table.select("1")
        self.assertIs(q.dialect, None)

    def test_table_with_dialect_query_cls(self):
        table = Table("abc", query_cls=SQLLiteQuery)
        q = table.select("1")
        self.assertIs(q.dialect, Dialects.SQLLITE)

    def test_table_factory_with_default_query_cls(self):
        table = Query.Table("abc")
        q = table.select("1")
        self.assertIs(q.dialect, None)

    def test_table_factory_with_dialect_query_cls(self):
        table = SQLLiteQuery.Table("abc")
        q = table.select("1")
        self.assertIs(q.dialect, Dialects.SQLLITE)

    def test_make_tables_factory_with_default_query_cls(self):
        t1, t2 = Query.Tables("abc", "def")
        q1 = t1.select("1")
        q2 = t2.select("2")
        self.assertIs(q1.dialect, None)
        self.assertIs(q2.dialect, None)

    def test_make_tables_factory_with_dialect_query_cls(self):
        t1, t2 = SQLLiteQuery.Tables("abc", "def")
        q1 = t1.select("1")
        q2 = t2.select("2")
        self.assertIs(q1.dialect, Dialects.SQLLITE)
        self.assertIs(q2.dialect, Dialects.SQLLITE)

    def test_table_with_bad_query_cls(self):
        with self.assertRaises(TypeError):
            Table("abc", query_cls=object)



import unittest

from pypika import (
    Tables,
)
from pypika import (
    functions as fn,
)
from pypika.terms import ValueWrapper


class NegationTests(unittest.TestCase):
    table_abc, table_efg = Tables("abc", "efg")

    def test_negate_wrapped_float(self):
        q = -ValueWrapper(1.0)

        self.assertEqual("-1.0", q.get_sql())

    def test_negate_wrapped_int(self):
        q = -ValueWrapper(1)

        self.assertEqual("-1", q.get_sql())

    def test_negate_field(self):
        q = -self.table_abc.foo

        self.assertEqual('-"abc"."foo"', q.get_sql(with_namespace=True, quote_char='"'))

    def test_negate_function(self):
        q = -fn.Sum(self.table_abc.foo)

        self.assertEqual('-SUM("abc"."foo")', q.get_sql(with_namespace=True, quote_char='"'))



from unittest import TestCase

from pypika import Field, Query, Table
from pypika.terms import AtTimezone


class FieldAliasTests(TestCase):
    t = Table("test", alias="crit")

    def test_when_alias_specified(self):
        c1 = Field("foo", alias="bar")
        self.assertEqual('bar', str(c1.alias))

        c1 = Field("foo").as_("bar")
        self.assertEqual('bar', str(c1.alias))


class FieldInitTests(TestCase):
    def test_init_with_str_table(self):
        test_table_name = "test_table"
        field = Field(name="name", table=test_table_name)
        self.assertEqual(field.table, Table(name=test_table_name))


class FieldHashingTests(TestCase):
    def test_tabled_eq_fields_equally_hashed(self):
        client_name1 = Field(name="name", table=Table("clients"))
        client_name2 = Field(name="name", table=Table("clients"))
        self.assertTrue(hash(client_name1) == hash(client_name2))

    def test_tabled_ne_fields_differently_hashed(self):
        customer_name = Field(name="name", table=Table("customers"))
        client_name = Field(name="name", table=Table("clients"))
        self.assertTrue(hash(customer_name) != hash(client_name))

    def test_non_tabled_aliased_eq_fields_equally_hashed(self):
        self.assertTrue(hash(Field(name="A", alias="my_a")) == hash(Field(name="A", alias="my_a")))

    def test_non_tabled_aliased_ne_fields_differently_hashed(self):
        self.assertTrue(hash(Field(name="A", alias="my_a1")) != hash(Field(name="A", alias="my_a2")))

    def test_non_tabled_eq_fields_equally_hashed(self):
        self.assertTrue(hash(Field(name="A")) == hash(Field(name="A")))

    def test_non_tabled_ne_fields_differently_hashed(self):
        self.assertTrue(hash(Field(name="A")) != hash(Field(name="B")))


class AtTimezoneTests(TestCase):
    def test_when_interval_not_specified(self):
        query = Query.from_("customers").select(AtTimezone("date", "US/Eastern"))
        self.assertEqual('SELECT "date" AT TIME ZONE \'US/Eastern\' FROM "customers"', str(query))

    def test_when_interval_specified(self):
        query = Query.from_("customers").select(AtTimezone("date", "-06:00", interval=True))
        self.assertEqual(
            'SELECT "date" AT TIME ZONE INTERVAL \'-06:00\' FROM "customers"',
            str(query),
        )

    def test_when_alias_specified(self):
        query = Query.from_("customers").select(AtTimezone("date", "US/Eastern", alias="alias1"))
        self.assertEqual(
            'SELECT "date" AT TIME ZONE \'US/Eastern\' "alias1" FROM "customers"',
            str(query),
        )

    def test_passes_kwargs_to_field_get_sql(self):
        customers = Table("customers")
        accounts = Table("accounts")
        query = (
            Query.from_(customers)
            .join(accounts)
            .on(customers.account_id == accounts.account_id)
            .select(AtTimezone(customers.date, "US/Eastern", alias="alias1"))
        )

        self.assertEqual(
            'SELECT "customers"."date" AT TIME ZONE \'US/Eastern\' "alias1" '
            'FROM "customers" JOIN "accounts" ON "customers"."account_id"="accounts"."account_id"',
            query.get_sql(with_namespace=True),
        )



import unittest

from pypika import CustomFunction, FunctionException, Query, Table

__author__ = "Airton Zanon"
__email__ = "me@airton.dev"


class TestUnitCustomFunction(unittest.TestCase):
    def test_should_fail_with_wrong_arguments(self):
        DateDiff = CustomFunction("DATE_DIFF", ["interval", "start_date", "end_date"])

        with self.assertRaises(FunctionException):
            DateDiff("foo")

    def test_should_return_function_with_arguments(self):
        DateDiff = CustomFunction("DATE_DIFF", ["interval", "start_date", "end_date"])

        self.assertEqual(
            "DATE_DIFF('day','start_date','end_date')",
            str(DateDiff("day", "start_date", "end_date")),
        )

    def test_should_return_function_with_no_arguments(self):
        CurrentDate = CustomFunction("CURRENT_DATE")

        self.assertEqual("CURRENT_DATE()", str(CurrentDate()))


class TestFunctionalCustomFunction(unittest.TestCase):
    def test_should_use_custom_function_on_select(self):
        service = Table("service")

        DateDiff = CustomFunction("DATE_DIFF", ["interval", "start_date", "end_date"])

        q = Query.from_(service).select(DateDiff("day", service.created_date, service.updated_date))

        self.assertEqual(
            'SELECT DATE_DIFF(\'day\',"created_date","updated_date") FROM "service"',
            str(q),
        )

    def test_should_fail_use_custom_function_on_select_with_wrong_arguments(self):
        service = Table("service")

        DateDiff = CustomFunction("DATE_DIFF", ["interval", "start_date", "end_date"])

        with self.assertRaises(FunctionException):
            Query.from_(service).select(DateDiff("day", service.created_date))
